""" Test for producing graphical outputs

    The report functionality is tested separately
    in test_integration

    Note
    ----

    Here we use the values returned from the plotted graph
    for testing. Regression tests against plotted graphs,
    as provided by image_comparison decorator of matplotlib,
    are not used since this is deprecated and also not consistent
    across different plotting engines.

"""


import os
import sys

import pytest

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

import pymrio  # noqa


@pytest.fixture()
def fix_testmrio_calc():
    """Single point to load the test mrio"""

    class TestMRIO:
        testmrio = pymrio.load_test().calc_all()

    return TestMRIO


def test_graphs_totals(fix_testmrio_calc):
    """Testing graph totals"""
    stressor = ("emission_type1", "air")
    tt = fix_testmrio_calc.testmrio
    ax = tt.emissions.plot_account(row=stressor)

    assert stressor[0] in str(ax.title)

    # In the plotting method data is extracted
    # for each account from all regions. Thus
    # the order of drawn patches is for each account
    # (cba, pba, imp, exp) for each region
    assert ax.patches[0].get_height() == tt.emissions.D_cba_reg.loc[stressor, "reg1"]
    assert ax.patches[1].get_height() == tt.emissions.D_cba_reg.loc[stressor, "reg2"]
    assert ax.patches[6].get_height() == tt.emissions.D_pba_reg.loc[stressor, "reg1"]

    with pytest.raises(ValueError):
        ax = tt.emissions.plot_account(row=stressor, per_capita=5)

    with pytest.raises(ValueError):
        ax = tt.emissions.plot_account(row=stressor, sector="food", per_capita=True)


def test_graphs_population_sector(fix_testmrio_calc):
    """Testing graph per population for a specific sector"""
    stressor = ("emission_type2", "water")
    sector = "mining"

    tt = fix_testmrio_calc.testmrio
    ax = tt.emissions.plot_account(
        row=stressor, per_capita=True, population=tt.population, sector=sector
    )

    assert stressor[0] in str(ax.title)
    assert stressor[1] in str(ax.title)
    assert sector in str(ax.title)
    assert "capita" in str(ax.title)

    # In the plotting method data is extracted
    # for each account from all regions. Thus
    # the order of drawn patches is for each account
    # (cba, pba, imp, exp) for each region
    sec_reg1_pop = (
        tt.emissions.D_cba.loc[stressor, ("reg1", sector)] / tt.population.reg1
    )[0]

    assert ax.patches[0].get_height() == sec_reg1_pop


def test_graphs_population_total(fix_testmrio_calc):
    """Testing total per population accounts"""
    stressor = ("emission_type2", "water")

    tt = fix_testmrio_calc.testmrio
    ax = tt.emissions.plot_account(row=stressor, per_capita=True)

    assert stressor[0] in str(ax.title)
    assert stressor[1] in str(ax.title)
    assert "capita" in str(ax.title)

    # In the plotting method data is extracted
    # for each account from all regions. Thus
    # the order of drawn patches is for each account
    # (cba, pba, imp, exp) for each region

    assert ax.patches[0].get_height() == tt.emissions.D_cba_cap.loc[stressor, "reg1"]
    assert ax.patches[1].get_height() == tt.emissions.D_cba_cap.loc[stressor, "reg2"]
    assert ax.patches[6].get_height() == tt.emissions.D_pba_cap.loc[stressor, "reg1"]



""" Tests the aggregation functionality in pymrio

This only test the top-level aggregation function.
For the low-level function 'build_agg_vec' and 'build_agg_matrix'
see test_util.py
"""

import os
import sys

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

import pymrio  # noqa


def test_aggreation_regions():
    """Testing aggregation of regions in various ways"""
    # All these representations should lead to the same results
    reg_agg_matrix = np.array([[1, 1, 1, 0, 0, 0], [0, 0, 0, 1, 1, 1]])

    reg_agg_vec1 = np.array([0, 0, 0, 1, 1, 1])
    reg_agg_vec2 = [0, 0, 0, 1, 1, 1]
    reg_agg_vec3 = ["a", "a", "a", "b", "b", "b"]

    reg_agg_df = pd.DataFrame(
        data=[
            ("reg1", "a"),
            ("reg5", "b"),
            ("reg3", "a"),
            ("reg4", "b"),
            ("reg2", "a"),
            ("reg6", "b"),
        ],
        index=range(6),
        columns=["original", "aggregated"],
    )

    io = pymrio.load_test()
    io.calc_all()
    manual_agg = (
        io.emissions.D_cba_reg.reg1
        + io.emissions.D_cba_reg.reg2
        + io.emissions.D_cba_reg.reg3
    )

    io_agg_wo_names = io.aggregate(region_agg=reg_agg_matrix, inplace=False)
    np.testing.assert_allclose(
        manual_agg.values, io_agg_wo_names.emissions.D_cba_reg.reg0
    )

    assert ["reg0", "reg1"] == io_agg_wo_names.get_regions().tolist()

    io_agg_with_names = io.aggregate(
        region_agg=reg_agg_matrix, region_names=["a", "b"], inplace=False
    )
    np.testing.assert_allclose(
        manual_agg.values, io_agg_with_names.emissions.D_cba_reg.a
    )

    assert ["a", "b"] == io_agg_with_names.get_regions().tolist()
    assert io_agg_with_names.unit.index.equals(io_agg_with_names.Z.index)

    io_vec1 = io.aggregate(region_agg=reg_agg_vec1, inplace=False)
    io_vec2 = io.aggregate(region_agg=reg_agg_vec2, inplace=False)
    io_vec3 = io.aggregate(region_agg=reg_agg_vec3, inplace=False)
    np.testing.assert_allclose(manual_agg.values, io_vec1.emissions.D_cba_reg.reg0)
    np.testing.assert_allclose(manual_agg.values, io_vec2.emissions.D_cba_reg.reg0)
    np.testing.assert_allclose(manual_agg.values, io_vec3.emissions.D_cba_reg.a)

    io_df = io.aggregate(region_agg=reg_agg_df, inplace=False)
    np.testing.assert_allclose(manual_agg.values, io_df.emissions.D_cba_reg.a)

    assert ["a", "b"] == io_df.get_regions().tolist()

    # Testing the aggregation of duplicate regions/sectors

    reg_rename_dict = reg_agg_df.set_index("original").squeeze().to_dict()
    io.rename_regions(reg_rename_dict)
    io_agg = io.aggregate_duplicates(inplace=False)
    io.aggregate_duplicates(inplace=True)

    pdt.assert_frame_equal(io.Z, io_agg.Z)
    pdt.assert_frame_equal(io.Z, io_agg_with_names.Z)
    pdt.assert_frame_equal(io.Y, io_agg.Y)
    pdt.assert_frame_equal(io.Y, io_agg_with_names.Y)
    pdt.assert_frame_equal(io.x, io_agg.x)
    pdt.assert_frame_equal(io.x, io_agg_with_names.x)

    pdt.assert_frame_equal(io.emissions.F, io_agg.emissions.F)
    pdt.assert_frame_equal(io.emissions.F, io_agg_with_names.emissions.F)
    pdt.assert_frame_equal(io.emissions.D_cba, io_agg.emissions.D_cba)
    pdt.assert_frame_equal(io.emissions.D_cba, io_agg_with_names.emissions.D_cba)
    pdt.assert_frame_equal(io.emissions.F_Y, io_agg.emissions.F_Y)
    pdt.assert_frame_equal(io.emissions.F_Y, io_agg_with_names.emissions.F_Y)

    pdt.assert_frame_equal(io.factor_inputs.F, io_agg.factor_inputs.F)
    pdt.assert_frame_equal(io.factor_inputs.F, io_agg_with_names.factor_inputs.F)
    pdt.assert_frame_equal(io.factor_inputs.D_cba, io_agg.factor_inputs.D_cba)
    pdt.assert_frame_equal(
        io.factor_inputs.D_cba, io_agg_with_names.factor_inputs.D_cba
    )

    assert io.unit.index.equals(io.Z.index)
    assert io_agg.unit.index.equals(io_agg.Z.index)


def test_aggreation_sectors():
    """Test different possibilities to aggregate sectors"""

    sec_agg_df = pd.DataFrame(
        data=[
            ("food", "eat"),
            ("mining", "dig"),
            ("manufactoring", "build"),
            ("electricity", "util"),
            ("construction", "build"),
            ("trade", "marg"),
            ("transport", "marg"),
            ("other", "misc"),
        ],
        columns=["original", "aggregated"],
    )
    io = pymrio.load_test()
    io_agg = io.aggregate(sector_agg=sec_agg_df, inplace=False)

    test_sectors = ["manufactoring", "construction"]
    assert (
        io.Z.loc["reg2", "reg2"].loc[test_sectors, test_sectors].sum().sum()
        == io_agg.Z.loc["reg2", "reg2"].loc["build", "build"]
    )
    assert (
        io.Y.loc["reg2", "reg2"].loc[test_sectors, :].sum().sum()
        == io_agg.Y.loc["reg2", "reg2"].loc["build", :].sum().sum()
    )

    test_rename = [("reg3", "other"), ("reg3", "misc")]
    assert (
        io.Z.loc[test_rename[0], test_rename[0]]
        == io_agg.Z.loc[test_rename[1], test_rename[1]]
    )

    sec_rename_dict = sec_agg_df.set_index("original").squeeze().to_dict()
    io.rename_regions(sec_rename_dict)
    io.aggregate_duplicates(inplace=True)

    io.calc_all()
    io_agg.calc_all()

    pdt.assert_frame_equal(io.Z, io_agg.Z)
    pdt.assert_frame_equal(io.Y, io_agg.Y)
    pdt.assert_frame_equal(io.x, io_agg.x)

    pdt.assert_frame_equal(io.emissions.F, io_agg.emissions.F)
    pdt.assert_frame_equal(io.emissions.D_cba, io_agg.emissions.D_cba)
    pdt.assert_frame_equal(io.emissions.F_Y, io_agg.emissions.F_Y)

    pdt.assert_frame_equal(io.factor_inputs.F, io_agg.factor_inputs.F)
    pdt.assert_frame_equal(io.factor_inputs.D_cba, io_agg.factor_inputs.D_cba)

    assert io.unit.index.equals(io.Z.index)
    assert io_agg.unit.index.equals(io_agg.Z.index)


def test_numerical_aggreation_sectors():
    """Testing aggregation of sectors with a numeric array"""
    sec_agg = np.array(
        [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 1, 1, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 1, 1]]
    )

    io = pymrio.load_test()
    io.calc_all()

    manual_agg = (
        io.emissions.D_pba.xs("trade", level="sector", axis=1)
        + io.emissions.D_pba.xs("transport", level="sector", axis=1)
        + io.emissions.D_pba.xs("other", level="sector", axis=1)
    )

    io.aggregate(sector_agg=sec_agg)

    np.testing.assert_allclose(
        manual_agg.values, io.emissions.D_pba.xs("sec2", level="sector", axis=1)
    )


def test_wrong_inputs():
    """Tests if correct Exceptions are raised for wrong shaped input arguments"""
    io = pymrio.load_test().calc_all()

    with pytest.raises(ValueError) as VA_sector_number:
        sec_agg = range(len(io.get_sectors()) - 1)
        _ = io.aggregate(sector_agg=sec_agg, inplace=False)
    assert "sector aggregation" in str(VA_sector_number.value).lower()

    with pytest.raises(ValueError) as VA_region_number:
        reg_agg = range(len(io.get_regions()) + 13)
        _ = io.aggregate(region_agg=reg_agg, inplace=False)
    assert "region aggregation" in str(VA_region_number.value).lower()

    with pytest.raises(ValueError) as VA_sector_name:
        sec_agg = range(len(io.get_sectors()))
        _ = io.aggregate(sector_agg=sec_agg, sector_names="C", inplace=False)
    assert "sector aggregation" in str(VA_sector_name.value).lower()

    with pytest.raises(ValueError) as VA_region_name:
        reg_agg = range(len(io.get_regions()))
        _ = io.aggregate(
            region_agg=reg_agg, region_names=["a", "b"], inplace=False  # noqa
        )
    assert "region aggregation" in str(VA_region_name.value).lower()


def test_total_agg():
    """Testing aggregation to total values"""
    io = pymrio.load_test().calc_all()
    np.testing.assert_allclose(
        io.emissions.D_cba.sum(axis=1).to_frame().values,
        io.aggregate(region_agg="global", sector_agg="total").emissions.D_cba.values,
    )


def test_underdefined_agg():
    """Testing correct error message for underdefined aggregation"""
    io = pymrio.load_test().calc_all()
    io.reset_all_to_coefficients()
    with pytest.raises(pymrio.core.mriosystem.AggregationError):
        io.aggregate(region_agg="global", sector_agg="total")



"""
Classification helpers for various standard MRIOs

KST 20211123
"""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from pymrio.core.constants import PYMRIO_PATH


def get_classification(mrio_name: Optional[str] = None):
    """Get predefined classifications included in pymrio

    Parameters
    ----------

    mrio_name: str
        MRIO for which to get the classification.
        Pass None (default) for a list of available classifications.

    Returns:
        pymrio.ClassificationData
    """

    if not mrio_name:
        return PYMRIO_PATH.keys()

    elif mrio_name not in PYMRIO_PATH.keys():
        raise ValueError(
            f"No classification available for {mrio_name}. Run the function without parameter to get available classifications."
        )

    return ClassificationData(
        mrio_classification_folder=PYMRIO_PATH[mrio_name],
        sectors_file_name="sectors.tsv",
        sectors_sep="\t",
        finaldemand_file_name="finaldemand.tsv",
        finaldemand_sep="\t",
    )


@dataclass
class ClassificationData(object):
    """Classification data for MRIOs

    Typically includes .sectors and .finaldemand

    """

    mrio_classification_folder: Union[str, Path]

    sectors_file_name: str
    sectors_sep: str

    finaldemand_file_name: str
    finaldemand_sep: str

    sectors: pd.DataFrame = field(init=False)
    finaldemand: pd.DataFrame = field(init=False)

    def __post_init__(self):
        self.sectors = pd.read_csv(
            Path(self.mrio_classification_folder) / self.sectors_file_name,
            sep=self.sectors_sep,
        )
        self.finaldemand = pd.read_csv(
            Path(self.mrio_classification_folder) / self.finaldemand_file_name,
            sep=self.finaldemand_sep,
        )

    def get_sector_dict(
        self, orig: Union[str, list, pd.Series], new: Union[str, pd.Series]
    ):
        """Returns sector rename dict based

        Parameters
        ----------
        orig: str or pd.Series
            Original classification, obtained by calling 'get_sectors()' on
            the given mrio or by using an instance of the class (which
            provides tab completion in an interactive environment) or by
            giving the column header name used in the classification file

        new: str or pd.Series
            New classification, name as given in the classification file. Will
            be extracted from the Series name if this is given (to allow for
            tab completion).

        Returns
        --------
            dict
        """
        if type(orig) is not str:
            for cname, cvalue in self.sectors.items():
                if cvalue.isin(orig).all():
                    orig = cname
                    break
            else:
                raise ValueError("No fully matching sector classification found")

        new = new if type(new) is str else new.name
        return self.sectors.loc[:, [orig, new]].set_index(orig).squeeze().to_dict()



""" Tests the parsing of different MRIOs """

import os
import sys

import numpy as np
import pandas.testing as pdt
import pytest

try:
    testpath = os.path.dirname(os.path.abspath(__file__))
except NameError:
    testpath = os.getcwd()  # easy run in interpreter
    if "tests" not in testpath:
        testpath = os.path.join(testpath, "tests")


sys.path.append(os.path.join(testpath, ".."))

import pymrio  # noqa


@pytest.fixture()
def fix_testmrio_calc():
    """Single point to load the test mrio"""

    class TestMRIO:
        testmrio = pymrio.load_test().calc_all()

    return TestMRIO


def test_parse_exio1(fix_testmrio_calc):
    exio1_mockpath = os.path.join(testpath, "mock_mrios", "exio1_mock")

    test_mrio = fix_testmrio_calc.testmrio

    exio1 = pymrio.parse_exiobase1(exio1_mockpath)

    exio1.calc_all()

    assert test_mrio.emissions.S.iloc[1, 1] == pytest.approx(
        exio1.emissions.S.iloc[1, 1]
    )

    assert test_mrio.emissions.D_cba.iloc[0, 5] == pytest.approx(
        exio1.emissions.D_cba.iloc[0, 5]
    )

    assert exio1.get_regions()[0] == "reg1"

    with pytest.raises(pymrio.ParserError):
        _ = pymrio.parse_exiobase1("foo")


def test_parse_exio2(fix_testmrio_calc):
    exio2_mockpath = os.path.join(
        testpath,
        "mock_mrios",
        "exio2_mock",
        "mrIOT_PxP_ita_coefficient_version2.2.2.zip",
    )

    test_mrio = fix_testmrio_calc.testmrio

    exio2 = pymrio.parse_exiobase2(exio2_mockpath, popvector=None)
    exio2.calc_all()

    assert test_mrio.emissions.S.iloc[1, 1] == pytest.approx(
        exio2.emissions.S.iloc[1, 1]
    )

    assert test_mrio.emissions.D_cba.iloc[0, 5] == pytest.approx(
        exio2.emissions.D_cba.iloc[0, 5]
    )

    pdt.assert_series_equal(
        exio2.impact.D_cba.loc["total emissions"],
        test_mrio.emissions.D_cba.sum(axis=0),
        check_names=False,
    )

    assert exio2.get_regions()[0] == "reg1"

    with pytest.raises(pymrio.ParserError):
        _ = pymrio.parse_exiobase2("foo")

    with pytest.raises(pymrio.ParserError):
        _ = pymrio.parse_exiobase2("foo.zip")


def test_parse_exio3(fix_testmrio_calc):
    exio3_mockpath = os.path.join(testpath, "mock_mrios", "exio3_mock")

    test_mrio = fix_testmrio_calc.testmrio

    exio3 = pymrio.parse_exiobase3(exio3_mockpath)
    exio3.calc_all()

    assert np.allclose(exio3.A, test_mrio.A)
    assert np.allclose(exio3.satellite.D_cba, test_mrio.emissions.D_cba)

    # Test the renaming of the regions - depends on the correct pseudo
    # naming in the mock mrio
    assert list(exio3.get_regions()) == ["AU", "BR", "FI", "FR", "KR", "WF"]


def test_parse_exio_ext():
    ext_mockpath = os.path.join(testpath, "mock_mrios", "exio_ext_mock")

    col5 = os.path.join(ext_mockpath, "ext_5col.txt")
    col3 = os.path.join(ext_mockpath, "ext_3col.txt")
    col2 = os.path.join(ext_mockpath, "ext_2col.txt")
    col1 = os.path.join(ext_mockpath, "ext_1col.txt")

    ext5 = pymrio.parse_exio12_ext(
        col5, index_col=5, name="col5", sep="\t", drop_compartment=True
    )
    ext3 = pymrio.parse_exio12_ext(
        col3, index_col=3, name="col3", sep="\t", drop_compartment=True
    )
    ext2 = pymrio.parse_exio12_ext(
        col2, index_col=2, name="col2", sep="\t", drop_compartment=True
    )
    ext1 = pymrio.parse_exio12_ext(
        col1, index_col=1, name="col1", sep="\t", drop_compartment=True
    )
    ext5_wcmp = pymrio.parse_exio12_ext(
        col5, index_col=5, name="col5", sep="\t", drop_compartment=False
    )
    ext3_wcmp = pymrio.parse_exio12_ext(
        col3, index_col=3, name="col3", sep="\t", drop_compartment=False
    )
    ext2_wcmp = pymrio.parse_exio12_ext(
        col2, index_col=2, name="col2", sep="\t", drop_compartment=False
    )
    ext1_wcmp = pymrio.parse_exio12_ext(
        col1, index_col=1, name="col1", sep="\t", drop_compartment=False
    )

    assert np.allclose(ext5.F, ext3.F, ext2.F, ext1.F, ext5_wcmp.F)

    assert ext5.F.iloc[1, 1] == 102
    assert ext5.F.iloc[-1, -1] == 148

    assert ext2.name == "col2"

    assert ext1.unit is None
    assert ext1_wcmp.unit is None
    assert ext5_wcmp.unit.iloc[0, 0] == "kg"
    assert ext3.unit.iloc[1, 0] == "kg"

    assert ext5.F.index.name == "stressor"
    assert list(ext5_wcmp.F.index.names) == ["stressor", "compartment"]
    assert list(ext3_wcmp.F.index.names) == ["stressor", "compartment"]
    assert ext2_wcmp.F.index.name == "stressor"


@pytest.mark.filterwarnings("ignore:Extension data")
def test_parse_wiod():
    wiod_mockpath = os.path.join(testpath, "mock_mrios", "wiod_mock")
    ww_path = pymrio.parse_wiod(path=wiod_mockpath, year=2009)
    ww_file = pymrio.parse_wiod(
        path=os.path.join(wiod_mockpath, "wiot09_row_sep12.xlsx")
    )

    ww_path.calc_all()
    ww_file.calc_all()

    pdt.assert_frame_equal(ww_path.Z, ww_file.Z)
    pdt.assert_frame_equal(ww_path.lan.D_cba, ww_file.lan.D_cba)
    pdt.assert_frame_equal(ww_path.SEA.D_pba, ww_file.SEA.D_pba)
    pdt.assert_frame_equal(ww_path.AIR.D_imp, ww_file.AIR.D_imp)

    assert ww_file.AIR.F_Y.loc["NOX", ("LUX", "c37")] == 999
    assert ww_file.SEA.F.loc["EMP", ("RoW", "19")] == 0


def test_oecd_2016():
    oecd_mockpath = os.path.join(testpath, "mock_mrios", "oecd_mock")
    oecd_IO_file = os.path.join(oecd_mockpath, "ICIO2016_2003.csv")

    oecd_file = pymrio.parse_oecd(path=oecd_IO_file)
    oecd_year = pymrio.parse_oecd(path=oecd_mockpath, year=2003)

    assert oecd_file == oecd_year

    with pytest.raises(FileNotFoundError):
        _ = pymrio.parse_oecd(oecd_mockpath, year=1077)
    with pytest.raises(pymrio.ParserError):
        _ = pymrio.parse_oecd(oecd_mockpath)

    oecd = oecd_file
    oecd.calc_all()

    # Test standard values
    assert np.allclose(
        oecd.Z.loc[("AUS", "C01T05AGR"), ("AUS", "C01T05AGR")], 23697.221
    )
    assert np.allclose(oecd.Z.loc[("NZL", "C23PET"), ("AUS", "C20WOD")], 684.49784)
    assert np.allclose(oecd.Y.loc[("NZL", "C23PET"), ("PER", "DIRP")], 14822221.0)

    # Test aggregation
    assert "CHN" in oecd.get_regions()
    assert "CN1" not in oecd.get_regions()
    assert "CN2" not in oecd.get_regions()

    assert oecd.Z.shape == (32, 32)

    assert np.allclose(
        oecd.Z.loc[("CHN", "C01T05AGR"), ("AUS", "C15T16FOD")], 6932.1858
    )
    assert np.allclose(
        oecd.Z.loc[("CHN", "C01T05AGR"), ("AUS", "C15T16FOD")], 6932.1858
    )
    assert np.allclose(oecd.Y.loc[("CHN", "C23PET"), ("NZL", "HFCE")], 24074818)
    assert np.allclose(
        oecd.factor_inputs.F.loc["VA+TAXSUB", ("CHN", "C20WOD")], 569612.84
    )


def test_oecd_2018():
    oecd_mockpath = os.path.join(testpath, "mock_mrios", "oecd_mock")
    oecd_IO_csv = os.path.join(oecd_mockpath, "ICIO2018_2010.CSV")
    oecd_IO_zip = os.path.join(oecd_mockpath, "ICIO2018_2010.zip")

    oecd_csv = pymrio.parse_oecd(path=oecd_IO_csv)
    oecd_zip = pymrio.parse_oecd(path=oecd_IO_zip)
    oecd_year = pymrio.parse_oecd(path=oecd_mockpath, year=2010)

    assert oecd_csv == oecd_zip == oecd_year

    oecd = oecd_zip

    # Test aggregation
    assert "MEX" in oecd.get_regions()
    assert "MX1" not in oecd.get_regions()

    assert np.allclose(oecd.Z.loc[("BEL", "16"), ("AUS", "05T06")], 150.89421)
    assert np.allclose(oecd.Y.loc[("BRA", "09"), ("COL", "GGFC")], 11211977)
    assert np.allclose(oecd.factor_inputs.F.loc["VALU", ("MEX", "01T03")], 284552.21)
    assert np.allclose(
        oecd.factor_inputs.F.loc["MEX_TAXSUB", ("BEL", "16")], 1.66161232097811
    )


def test_parse_eora26(fix_testmrio_calc):
    eora_mockpath = os.path.join(testpath, "mock_mrios", "eora26_mock")

    test_mrio = fix_testmrio_calc.testmrio

    eora_short = pymrio.parse_eora26(eora_mockpath, year=2010)
    eora_full = pymrio.parse_eora26(eora_mockpath, year=2010, country_names="full")

    eora_short.calc_all()

    assert test_mrio.emissions.D_cba.iloc[1, 1] == pytest.approx(
        eora_short.Q.D_cba.iloc[1, 1]
    )

    assert eora_short.get_regions()[0] == "reg1"
    assert eora_full.get_regions()[0] == "Region 1"

    with pytest.raises(pymrio.ParserError):
        _ = pymrio.parse_eora26(eora_mockpath, year=2010, country_names="bogus")


if __name__ == "__main__":
    test_oecd_2016()



"""
Various constant values

KST 20140903
"""

import os
from pathlib import Path

# path information of the package
__ROOT = Path(os.path.dirname(__file__))
PYMRIO_PATH = {
    "test_mrio": Path(__ROOT / "../mrio_models/test_mrio").absolute(),
    "test_mrio_data": Path(__ROOT / "../mrio_models/test_mrio/mrio_data").absolute(),
    "testmrio": Path(__ROOT / "../mrio_models/test_mrio").absolute(),
    "test": Path(__ROOT / "../mrio_models/test_mrio").absolute(),
    # TODO remove "exio2": just here to keep the tests running
    "exio2": Path(__ROOT / "../mrio_models/exio2_pxp").absolute(),
    "exio2_pxp": Path(__ROOT / "../mrio_models/exio2_pxp").absolute(),
    "exio2_ixi": Path(__ROOT / "../mrio_models/exio2_ixi").absolute(),
    "exio3_pxp": Path(__ROOT / "../mrio_models/exio3_pxp").absolute(),
    "exio3_ixi": Path(__ROOT / "../mrio_models/exio3_ixi").absolute(),
}

# generic names (needed for the aggregation  if no names are given)
GENERIC_NAMES = {
    "sector": "sec",
    "region": "reg",
    "iosys": "IOSystem",
    "ext": "Extension",
}

DEFAULT_FILE_NAMES = {
    "filepara": "file_parameters.json",
    "metadata": "metadata.json",
    "download_log": "download_log.json",
}

MISSING_AGG_ENTRY = {
    "region": "Unspecified region",
    "sector": "Unspecified sector",
}

STORAGE_FORMAT = {
    "txt": ["txt", "tsv", "csv"],
    "parquet": ["parquet", "par", "parq"],
    "pickle": ["pickle", "pkl"],
}


#  Download links for Gloria MRIO files
GLORIA_URLS = {
    "053": [
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACPnp0qOD1N7CSjv0reFKSba/previous_releases/053/MRIO/GLORIA_MRIOs_53_2006.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACD6UmIBqbmkAQiAOfX7U_fa/previous_releases/053/MRIO/GLORIA_MRIOs_53_1997.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD6y_13ul9SZnJkPV1GhZCza/previous_releases/053/MRIO/GLORIA_MRIOs_53_2013.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAA5bFH6uzjqhm6fowdMlkyLa/previous_releases/053/MRIO/GLORIA_MRIOs_53_1996.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB3PLbnxPHIP2MGN8S36ZUBa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2007.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABIfGIVZkelhq58Ij1DKO1Va/previous_releases/053/MRIO/GLORIA_MRIOs_53_2012.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAqOhcAo8jFU_GhTu1fzOKWa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2010.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABspySp6y4Jh2OaLrTlSeY_a/previous_releases/053/MRIO/GLORIA_MRIOs_53_1994.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABAXczzTGbWj8i-ho0fSTsLa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2005.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABK94uXvjIdWzrgCP5uxk6Wa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2019.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABTfP3rpj_QkIPyPIJqVo62a/previous_releases/053/MRIO/GLORIA_MRIOs_53_2018.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABW9BpSx6h83gpmyCnVfKZ-a/previous_releases/053/MRIO/GLORIA_MRIOs_53_2011.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADvhpsqMor0wvNFLD-zWcRAa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2004.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADS3ffwpUxjkoLlQneJrrUDa/previous_releases/053/MRIO/GLORIA_MRIOs_53_1995.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABdMfXMCF6b3HqTJEjyQkTsa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2009.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAfHBfF-5N0ZaYr-gwfJKSya/previous_releases/053/MRIO/GLORIA_MRIOs_53_1998.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADWa9MT3qH_YAVwlqb2-eEBa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2015.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABT4WkwUQt_5KbPsWMgUUFba/previous_releases/053/MRIO/GLORIA_MRIOs_53_1991.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABfxaEEsKc8MKndscJMxilha/previous_releases/053/MRIO/GLORIA_MRIOs_53_2000.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB7j7bikNb1q3RhTyyLB9mQa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2014.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABrkCHIOdLU5zUmHuI8nDKma/previous_releases/053/MRIO/GLORIA_MRIOs_53_2001.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAW_BEqaeIvAzmFE6oYz8rKa/previous_releases/053/MRIO/GLORIA_MRIOs_53_1990.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB2ktuivgPwclQXVNQbh_Afa/previous_releases/053/MRIO/GLORIA_MRIOs_53_1999.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADyg36facQvwueC6r85yEY8a/previous_releases/053/MRIO/GLORIA_MRIOs_53_2008.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAA7Z0xEExKNK91GvinGGk9va/previous_releases/053/MRIO/GLORIA_MRIOs_53_2003.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD1rYcXCwCvNxC4wV8apUFCa/previous_releases/053/MRIO/GLORIA_MRIOs_53_1992.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD-J1n98SSYniRLSNfEe0kJa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2016.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAQc9jieTHGraE9nRGha1g4a/previous_releases/053/MRIO/GLORIA_MRIOs_53_1993.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAzRQsWiNc6Wu_5mOXAV1YIa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2002.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADPrsw4iWytQ9ItCpJ8ZqwIa/previous_releases/053/MRIO/GLORIA_MRIOs_53_2017.zip?dl=0",
    ],
    "054": [
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC7K5dfdnOA3I_CJm4Bu0nHa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2016.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACO5ce_FtnXN-w8ZykcqrCYa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2003.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACvnqqeofEQPAG0X-BWxNRta/previous_releases/054/MRIO/GLORIA_MRIOs_54_1992.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABXuHtPoMajYNR5YH18Frf_a/previous_releases/054/MRIO/GLORIA_MRIOs_54_2017.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACOfDR7Jtk83cEzFQwktbGja/previous_releases/054/MRIO/GLORIA_MRIOs_54_1993.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACmf_zQswJz60ZWzorQ6Bpua/previous_releases/054/MRIO/GLORIA_MRIOs_54_2002.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADXNVDIgSjiTq2XYZBOp3TXa/previous_releases/054/MRIO/GLORIA_MRIOs_54_1991.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABpxrh-7mFDL6va4d6Z3E_za/previous_releases/054/MRIO/GLORIA_MRIOs_54_2000.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACwD1DOvsiLNxsMreU3Vo3ua/previous_releases/054/MRIO/GLORIA_MRIOs_54_2015.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABlYAdTmli3iDfM3li2OAxxa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2009.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB-HUu421eE7VkH0yQV6KwAa/previous_releases/054/MRIO/GLORIA_MRIOs_54_1998.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACqNzqTjYXrR6cJ7fpp2-_ua/previous_releases/054/MRIO/GLORIA_MRIOs_54_1999.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABPAFDyayr8LUfTr1R8F6DAa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2008.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB1In5emdB928-a0B1j_Uo0a/previous_releases/054/MRIO/GLORIA_MRIOs_54_2001.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABHnhwWKGOYcSMyXFxCS7Dfa/previous_releases/054/MRIO/GLORIA_MRIOs_54_1990.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADZ2etvFvgEUOoCWRHgfUcNa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2014.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB5bz-aVrQVkj5ZSuAuGhuAa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2019.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACHEcV9TmcUUIXxd2trwzj0a/previous_releases/054/MRIO/GLORIA_MRIOs_54_1994.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAADEQMUbcrbfItvHEgs9Tqxa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2005.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAAC8oUPc0Bf_xpiS-vMeWNa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2010.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACiG4RqHNhtOTvNkqPGvFEsa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2004.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAA7I_SrhI8gklR7Za2drsNFa/previous_releases/054/MRIO/GLORIA_MRIOs_54_1995.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACmGlhuY9zeijQCdllbwkdpa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2011.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD8snAscs6LFBF4mRlo7qsfa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2018.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACr37Rp_r7nbxaUgv8TzvlHa/previous_releases/054/MRIO/GLORIA_MRIOs_54_2013.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAujo-1iMVSz7N3ozLLN3Zga/previous_releases/054/MRIO/GLORIA_MRIOs_54_2006.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAt3nE7WeWQxzkeVtabsm4Sa/previous_releases/054/MRIO/GLORIA_MRIOs_54_1997.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABFO3rlAXAI9MuZfd-KQ-9ia/previous_releases/054/MRIO/GLORIA_MRIOs_54_2012.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB4j8NBysXu4MkgvZYmrSwfa/previous_releases/054/MRIO/GLORIA_MRIOs_54_1996.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC-XyiKcDOn9TYmeDPKjyfna/previous_releases/054/MRIO/GLORIA_MRIOs_54_2007.zip?dl=0",
    ],
    "055": [
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACmYOJiT8mK1uHLQlCgn4Oya/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1990.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD11mTRGEtcqf1dRMFtQrYza/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1991.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAljBvDe2kgEL7AjDOPfF5Ca/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1992.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACdxKvPfZo8JzkQIJxbo3rxa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1993.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACFPwnkn4sWCdzmxXTXNLJKa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1994.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABnBRFa0IY4iS8yU3U3bQ-ca/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1995.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC1Ms5inScRNGD_UtqRZ4cba/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1996.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABzqn-iNshbGn3PV2rDgOEma/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1997.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAfXq94cK9HgqDGRw0Av51Ra/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1998.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABPCBRPcNkDgIKV0f4Aoh-aa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_1999.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABeYfN4oPwiHGMVvrdM2H7ia/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2000.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABbWD78TEiNbnZs5pZkMNSaa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2001.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADwlWrpZShpxWm0frKPU3Aja/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2002.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACFSmp0PcO7UeRmC8xokRwea/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2003.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAc6MD3-u6nvoPgg3cL9tvoa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2004.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC2cg37BrUzSg3pQg3CxnLDa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2005.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACF6A0mQQ9UwvsBKCU05z0va/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2006.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAATYbJBOR6XHN6H4QRAC6tFa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2007.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAJ-gud5eL1kgvJvXAT-3zja/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2008.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAA5EQmc0cDiVe6KfOsnZXesa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2009.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACTtJbqQWeSNVe1Z0jYJ_rra/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2010.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADttP2Z4KGdU4mnd58f0w2Sa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2011.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADS7y82yMnifEnpPHPuoxITa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2012.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACSkVqnHjhc1k9NY_3uDgJFa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2013.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADxMiz9payAd6cpgxoUDIw4a/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2014.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC9BjOyOaaSENX_DfMcIU-qa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2015.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD3y0caXYJPtvvZezhXEQUUa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2016.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAF9FaRBH7Hk1r48GYcilfGa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2017.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACyvsy2YPCAFqBgHC2YZWaba/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2018.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADSHS7zJ05U5Zm4FUtrldpRa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2019.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAA_hJiYUEo8gVZbqp7HCgMPa/previous_releases/055/GLORIA_MRIO_Loop055_part_I_MRIOdatabase/GLORIA_MRIOs_55_2020.zip?dl=0",
    ],
    "057": [
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAILpkD09rQ_bAeeWDZuE3aa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1990.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABt_nByH1FuML746fiALCEAa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1991.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACUtfp-ZS1SbjF7bdILg_ZAa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1992.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAW5l8EqE9_Ejreqb5_qLdUa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1993.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAU4AFSGoeYUJQhZC5_u_lha/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1994.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACK97tQE7V1OYpJ9wa7ab7da/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1995.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACXilhQsMeCgsupXjohYpEma/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1996.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB_C_8kESAFr4770dhhV51Ya/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1997.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABmQag5TZCeaouETVKbFsppa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1998.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABQ_-L9l_7wJWF50QTdOCZTa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_1999.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACh7azwKIUvyyOgTM4K5A9Ta/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2000.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAClJDe_V5KD6j-bq0ImrZda/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2001.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC1j_FPoic70F6G5Kuaw3fUa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2002.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAK11s6S112EKxUwWPgbHxVa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2003.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADC0tZBKH_kp5sxvkK_hQC8a/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2004.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAj-vEmgmrbfFEhN6rNi8D0a/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2005.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB0UtTXqf5xH2zqBxvrrA8za/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2006.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADw_eRYFWulRfLkQwFVmryNa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2007.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAC2346imGfZo88BZFwmIqEga/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2008.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAhfCAMfMyxOqHC4R5m1Mzqa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2009.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACaJ9PVcSxSS0lYjG01Wb0Ga/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2010.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADMPuCtsivFzTMcNhH9H15Ia/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2011.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABbw6LJJSPmuZGobw2vffpDa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2012.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACyvVf4ZaCtUJ6YfZsbITYwa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2013.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACposoILlF8OETb20yh4st7a/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2014.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAE0UCMywuaUk72q15STq7Qa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2015.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADOL4JYoO9WFKKij1h27m4fa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2016.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACyDpP4C4juFk7n1LOJyq5fa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2017.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACvB106Cw7vxUEiu_un5WYAa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2018.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAD_3uLb7NYSKnZhg6YyWSSna/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2019.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAdrXgf7Ie4K8wE3WxgMhKVa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2020.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACOZnEwhznnyneXanq_QEvia/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2021.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAAuKImX1cnJBAWzke3Mv4BRa/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2022.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AADN-i7oupC0I7hmQmXMwrSia/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2023.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AABt_-yvC5HxFclnWomiBmZ3a/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2024.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAB9JR7HbgkHPUtO5TwVcPvva/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2025.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AAASkSkOcuLYyx7J9Rby_xo1a/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2026.zip?dl=0",
        "https://dl.dropboxusercontent.com/sh/o4fxq94n7grvdbk/AACp-7VsPPfwlnjMOY3uZ2Gna/latest_release/057/GLORIA_MRIO_Loop057_part_I_MRIOdatabase/GLORIA_MRIOs_57_2027.zip?dl=0",
    ],
}



""" test cases for all mathematical functions """

import os
import sys

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))


# the function which should be tested here
from pymrio.tools.iomath import calc_A  # noqa
from pymrio.tools.iomath import calc_accounts  # noqa
from pymrio.tools.iomath import calc_e  # noqa
from pymrio.tools.iomath import calc_F  # noqa
from pymrio.tools.iomath import calc_F_Y  # noqa
from pymrio.tools.iomath import calc_gross_trade  # noqa
from pymrio.tools.iomath import calc_L  # noqa
from pymrio.tools.iomath import calc_M  # noqa
from pymrio.tools.iomath import calc_S  # noqa
from pymrio.tools.iomath import calc_S_Y  # noqa
from pymrio.tools.iomath import calc_x  # noqa
from pymrio.tools.iomath import calc_x_from_L  # noqa
from pymrio.tools.iomath import calc_Z  # noqa


# test data
@pytest.fixture()
def td_IO_Data_Miller():
    """This data is from the chapter 2 of
    Input-output analysis: foundations and extensions -- Miller, Ronald E
    and Blair, Peter D -- 2009 (ISBN: 9780521517133)
    """

    class IO_Data_Miller:
        # Table 2.3 of the Book
        Z_arr = np.array([[150, 500], [200, 100]]).astype("float")
        Z_df = pd.DataFrame(
            data=Z_arr,
            index=["sec1", "sec2"],
            columns=["sec1", "sec2"],
        )

        # Table 2.4 of the Book
        A_arr = np.array([[0.15, 0.25], [0.20, 0.05]]).astype("float")
        A_df = pd.DataFrame(
            data=A_arr,
            index=["sec1", "sec2"],
            columns=["sec1", "sec2"],
        )

        # Table 2.3 of the Book
        fd_arr = np.array([[350], [1700]]).astype("float")
        fd_df = pd.DataFrame(
            data=fd_arr,
            index=["sec1", "sec2"],
            columns=["fi"],
        )

        # Table 2.3 of the Book
        x_arr = np.array([[1000], [2000]]).astype("float")
        x_df = pd.DataFrame(
            data=x_arr,
            index=["sec1", "sec2"],
            columns=["indout"],
        )

        # At the example following Table 2.3, additional
        # decimals where calculated
        L_arr = np.array([[1.25412541, 0.330033], [0.2640264, 1.12211221]]).astype(
            "float"
        )
        L_df = pd.DataFrame(
            data=L_arr,
            index=["sec1", "sec2"],
            columns=["sec1", "sec2"],
        )

        # Example epsilon from chapter 2, it uses the new x
        xnew_arr = np.array([[1247.5], [1841.6]]).astype("float")
        xnew_df = pd.DataFrame(
            data=xnew_arr,
            index=["sec1", "sec2"],
            columns=["indout"],
        )

        labcoeff_arr = np.array(
            [
                [0.3, 0.25],
            ]
        )
        labcoeff_df = pd.DataFrame(
            data=labcoeff_arr,
            index=["total labor"],
            columns=["sec1", "sec2"],
        )

        labtot_arr = np.array(
            [
                [374.25, 460.4],
            ]
        )  # .25 because of rounding
        labtot_df = pd.DataFrame(
            data=labtot_arr,
            index=["total labor"],
            columns=["sec1", "sec2"],
        )

    return IO_Data_Miller


@pytest.fixture()
def td_small_MRIO():
    """A small MRIO with three sectors and two regions.
    The testdata here just consists of pandas DataFrames, the functionality
    with numpy arrays gets tested with td_IO_Data_Miller.
    """

    class IO_Data:
        _sectors = ["sector1", "sector2", "sector3"]
        _regions = ["reg1", "reg2"]
        _Z_multiindex = pd.MultiIndex.from_product(
            [_regions, _sectors], names=["region", "sector"]
        )

        Z = pd.DataFrame(
            data=[
                [10, 5, 1, 6, 5, 7],
                [0, 2, 0, 0, 5, 3],
                [10, 3, 20, 4, 2, 0],
                [5, 0, 0, 1, 10, 9],
                [0, 10, 1, 0, 20, 1],
                [5, 0, 0, 1, 10, 10],
            ],
            index=_Z_multiindex,
            columns=_Z_multiindex,
            dtype=("float64"),
        )

        _categories = ["final demand"]
        _Y_multiindex = pd.MultiIndex.from_product(
            [_regions, _categories], names=["region", "category"]
        )
        Y = pd.DataFrame(
            data=[[14, 3], [2.5, 2.5], [13, 6], [5, 20], [10, 10], [3, 10]],
            index=_Z_multiindex,
            columns=_Y_multiindex,
            dtype=("float64"),
        )

        F = pd.DataFrame(
            data=[[20, 1, 42, 4, 20, 5], [5, 4, 11, 8, 2, 10]],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
            dtype=("float64"),
        )

        F_Y = pd.DataFrame(
            data=[[50, 10], [100, 20]],
            index=["ext_type_1", "ext_type_2"],
            columns=_Y_multiindex,
            dtype=("float64"),
        )

        S_Y = pd.DataFrame(
            data=[
                [1.0526315789473684, 0.1941747572815534],
                [2.1052631578947367, 0.3883495145631068],
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Y_multiindex,
            dtype=("float64"),
        )

        A = pd.DataFrame(
            data=[
                [
                    0.19607843137254902,
                    0.3333333333333333,
                    0.017241379310344827,
                    0.12,
                    0.09615384615384616,
                    0.1794871794871795,
                ],  # noqa
                [
                    0.0,
                    0.13333333333333333,
                    0.0,
                    0.0,
                    0.09615384615384616,
                    0.07692307692307693,
                ],  # noqa
                [
                    0.19607843137254902,
                    0.2,
                    0.3448275862068966,
                    0.08,
                    0.038461538461538464,
                    0.0,
                ],  # noqa
                [
                    0.09803921568627451,
                    0.0,
                    0.0,
                    0.02,
                    0.19230769230769232,
                    0.23076923076923075,
                ],  # noqa
                [
                    0.0,
                    0.6666666666666666,
                    0.017241379310344827,
                    0.0,
                    0.38461538461538464,
                    0.02564102564102564,
                ],  # noqa
                [
                    0.09803921568627451,
                    0.0,
                    0.0,
                    0.02,
                    0.19230769230769232,
                    0.2564102564102564,
                ],  # noqa
            ],
            index=_Z_multiindex,
            columns=_Z_multiindex,
        )

        L = pd.DataFrame(
            data=[
                [
                    1.3387146304736708,
                    0.9689762471208287,
                    0.05036622549592462,
                    0.17820960407435948,
                    0.5752019383714646,
                    0.4985179148178926,
                ],  # noqa
                [
                    0.02200779585580331,
                    1.3716472861392823,
                    0.0076800357678581885,
                    0.006557415453762468,
                    0.2698335633228079,
                    0.15854643902810828,
                ],  # noqa
                [
                    0.43290422861412026,
                    0.8627066565439678,
                    1.5492942759220427,
                    0.18491657196329184,
                    0.44027825642348534,
                    0.26630955082840885,
                ],  # noqa
                [
                    0.18799498787612925,
                    0.5244084722329316,
                    0.020254008037620782,
                    1.0542007368783255,
                    0.5816573175534603,
                    0.44685014763069275,
                ],  # noqa
                [
                    0.04400982046095892,
                    1.5325472495862535,
                    0.05259311578831879,
                    0.014602513642445088,
                    1.9545285794951548,
                    0.2410917825607805,
                ],  # noqa
                [
                    0.19294222439918532,
                    0.5382086951864299,
                    0.020787008249137116,
                    0.05562707205933412,
                    0.596964089068025,
                    1.4849251515157111,
                ],  # noqa
            ],
            index=_Z_multiindex,
            columns=_Z_multiindex,
        )

        x = pd.DataFrame(
            data=[
                [51],
                [15],
                [58],
                [50],
                [52],
                [39],
            ],
            columns=["indout"],
            index=_Z_multiindex,
            dtype=("float64"),
        )
        S = pd.DataFrame(
            data=[
                [
                    0.39215686274509803,
                    0.06666666666666667,
                    0.7241379310344828,
                    0.08,
                    0.38461538461538464,
                    0.1282051282051282,
                ],  # noqa
                [
                    0.09803921568627451,
                    0.26666666666666666,
                    0.1896551724137931,
                    0.16,
                    0.038461538461538464,
                    0.2564102564102564,
                ],  # noqa
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
        )

        M = pd.DataFrame(
            data=[
                [
                    0.896638324101429,
                    1.7965474933011258,
                    1.1666796580507097,
                    0.3013124703668342,
                    1.4371886818255364,
                    0.7177624711634629,
                ],  # noqa
                [
                    0.3004620527704837,
                    0.9052387706892087,
                    0.311411003349379,
                    0.23778766312079253,
                    0.5331560745059157,
                    0.6031791628984159,
                ],  # noqa
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
        )

        D_cba = pd.DataFrame(
            data=[
                [
                    14.059498889254174,
                    18.863255551508175,
                    17.32012296814962,
                    8.71616437964097,
                    18.863255551508175,
                    14.17770265993889,
                ],  # noqa
                [
                    5.395407054390734,
                    7.594657671782178,
                    5.857880532237175,
                    5.657139420727301,
                    7.594657671782178,
                    7.900257649080434,
                ],  # noqa
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
        )

        D_pba = pd.DataFrame(
            data=[
                [20.000000000000004, 1.0, 42.00000000000001, 4.0, 20.0, 5.0],  # noqa
                [5.000000000000001, 4.0, 11.0, 8.0, 1.9999999999999998, 10.0],  # noqa
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
        )

        D_imp = pd.DataFrame(
            data=[
                [
                    1.2792573306446737,
                    10.499069649181388,
                    1.2752266807875912,
                    6.604374747509535,
                    8.364185902326788,
                    10.842115601865368,
                ],  # noqa
                [
                    2.054905005956757,
                    3.9151998960771452,
                    1.522271392125981,
                    1.743464577633809,
                    3.6794577757050337,
                    3.221508682410955,
                ],  # noqa
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
        )

        D_exp = pd.DataFrame(
            data=[
                [
                    8.251832343364468,
                    0.5304113433404783,
                    17.02843256499675,
                    1.3307504334524414,
                    9.797226856351275,
                    1.9255763708099365,
                ],  # noqa
                [
                    2.062958085841117,
                    2.1216453733619134,
                    4.459827576546767,
                    2.6615008669048827,
                    0.9797226856351275,
                    3.851152741619873,
                ],  # noqa
            ],
            index=["ext_type_1", "ext_type_2"],
            columns=_Z_multiindex,
        )

    return IO_Data


def test_calc_x_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(
        td_IO_Data_Miller.x_df, calc_x(td_IO_Data_Miller.Z_df, td_IO_Data_Miller.fd_df)
    )


def test_calc_x_arr(td_IO_Data_Miller):
    npt.assert_array_equal(
        td_IO_Data_Miller.x_arr,
        calc_x(td_IO_Data_Miller.Z_arr, td_IO_Data_Miller.fd_arr),
    )


def test_calc_Z_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(
        td_IO_Data_Miller.Z_df, calc_Z(td_IO_Data_Miller.A_df, td_IO_Data_Miller.x_df)
    )


def test_calc_Z_arr(td_IO_Data_Miller):
    npt.assert_array_equal(
        td_IO_Data_Miller.Z_arr,
        calc_Z(td_IO_Data_Miller.A_arr, td_IO_Data_Miller.x_arr),
    )


def test_calc_A_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(
        td_IO_Data_Miller.A_df, calc_A(td_IO_Data_Miller.Z_df, td_IO_Data_Miller.x_df)
    )


def test_calc_A_arr(td_IO_Data_Miller):
    npt.assert_array_equal(
        td_IO_Data_Miller.A_arr,
        calc_A(td_IO_Data_Miller.Z_arr, td_IO_Data_Miller.x_arr),
    )


def test_calc_L_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(td_IO_Data_Miller.L_df, calc_L(td_IO_Data_Miller.A_df))


def test_calc_L_arr(td_IO_Data_Miller):
    npt.assert_allclose(
        td_IO_Data_Miller.L_arr, calc_L(td_IO_Data_Miller.A_arr), rtol=1e-5
    )


def test_calc_x_from_L_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(
        td_IO_Data_Miller.x_df,
        calc_x_from_L(td_IO_Data_Miller.L_df, td_IO_Data_Miller.fd_df),
    )


def test_calc_x_from_L_arr(td_IO_Data_Miller):
    npt.assert_allclose(
        td_IO_Data_Miller.x_arr,
        calc_x_from_L(td_IO_Data_Miller.L_arr, td_IO_Data_Miller.fd_arr),
        rtol=1e-5,
    )


def test_calc_F_arr(td_IO_Data_Miller):
    npt.assert_allclose(
        td_IO_Data_Miller.labtot_arr,
        calc_F(td_IO_Data_Miller.labcoeff_arr, td_IO_Data_Miller.xnew_arr),
        rtol=1e-5,
    )


def test_calc_F_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(
        td_IO_Data_Miller.labtot_df,
        calc_F(td_IO_Data_Miller.labcoeff_df, td_IO_Data_Miller.xnew_df),
    )


def test_calc_S_arr(td_IO_Data_Miller):
    npt.assert_allclose(
        td_IO_Data_Miller.labcoeff_arr,
        calc_S(td_IO_Data_Miller.labtot_arr, td_IO_Data_Miller.xnew_arr),
        rtol=1e-5,
    )


def test_calc_S_df(td_IO_Data_Miller):
    pdt.assert_frame_equal(
        td_IO_Data_Miller.labcoeff_df,
        calc_S(td_IO_Data_Miller.labtot_df, td_IO_Data_Miller.xnew_df),
    )


def test_calc_x_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(td_small_MRIO.x, calc_x(td_small_MRIO.Z, td_small_MRIO.Y))


def test_calc_A_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(td_small_MRIO.A, calc_A(td_small_MRIO.Z, td_small_MRIO.x))
    # we also test the different methods to provide x:
    x_values = td_small_MRIO.x.values
    x_Tvalues = td_small_MRIO.x.T.values
    x_series = pd.Series(td_small_MRIO.x.iloc[:, 0])
    pdt.assert_frame_equal(td_small_MRIO.A, calc_A(td_small_MRIO.Z, x_series))
    pdt.assert_frame_equal(td_small_MRIO.A, calc_A(td_small_MRIO.Z, x_values))
    pdt.assert_frame_equal(td_small_MRIO.A, calc_A(td_small_MRIO.Z, x_Tvalues))


def test_calc_Z_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(td_small_MRIO.Z, calc_Z(td_small_MRIO.A, td_small_MRIO.x))
    # we also test the different methods to provide x:
    x_values = td_small_MRIO.x.values
    x_Tvalues = td_small_MRIO.x.T.values
    x_series = pd.Series(td_small_MRIO.x.iloc[:, 0])
    pdt.assert_frame_equal(td_small_MRIO.Z, calc_Z(td_small_MRIO.A, x_series))
    pdt.assert_frame_equal(td_small_MRIO.Z, calc_Z(td_small_MRIO.A, x_values))
    pdt.assert_frame_equal(td_small_MRIO.Z, calc_Z(td_small_MRIO.A, x_Tvalues))


def test_calc_L_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(td_small_MRIO.L, calc_L(td_small_MRIO.A))


def test_calc_S_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(td_small_MRIO.S, calc_S(td_small_MRIO.F, td_small_MRIO.x))


def test_calc_S_Y_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(
        td_small_MRIO.S_Y, calc_S_Y(td_small_MRIO.F_Y, td_small_MRIO.Y.sum(axis=0))
    )


def test_calc_F_Y_MRIO(td_small_MRIO):
    S_Y = calc_S_Y(td_small_MRIO.F_Y, td_small_MRIO.Y.sum(axis=0))
    pdt.assert_frame_equal(
        td_small_MRIO.F_Y, calc_F_Y(S_Y, td_small_MRIO.Y.sum(axis=0))
    )


def test_calc_M_MRIO(td_small_MRIO):
    pdt.assert_frame_equal(td_small_MRIO.M, calc_M(td_small_MRIO.S, td_small_MRIO.L))


def test_calc_gross_trade_MRIO(td_small_MRIO):
    gt = calc_gross_trade(td_small_MRIO.Z, td_small_MRIO.Y)

    reg1sec2trade = (
        td_small_MRIO.Z.loc[("reg1", "sector2"), "reg2"].sum()
        + td_small_MRIO.Y.loc[("reg1", "sector2"), "reg2"].sum()
    )

    total_exports_reg1 = td_small_MRIO.Z.loc["reg1", "reg2"].sum(
        axis=1
    ) + td_small_MRIO.Y.loc["reg1", "reg2"].sum(axis=1)

    assert gt.bilat_flows.loc[("reg1", "sector2"), "reg2"] == reg1sec2trade
    pdt.assert_series_equal(
        gt.totals.loc["reg1", "exports"],
        total_exports_reg1,
        check_names=False,
    )


def test_calc_accounts_MRIO(td_small_MRIO):
    # calc the accounts
    nD_cba, nD_pba, nD_imp, nD_exp = calc_accounts(
        td_small_MRIO.S,
        td_small_MRIO.L,
        td_small_MRIO.Y.groupby(level="region", axis=1, sort=False).sum(),
    )
    # test all

    pdt.assert_frame_equal(
        td_small_MRIO.D_cba,
        nD_cba,
    )
    pdt.assert_frame_equal(
        td_small_MRIO.D_pba,
        nD_pba,
    )
    pdt.assert_frame_equal(
        td_small_MRIO.D_imp,
        nD_imp,
    )
    pdt.assert_frame_equal(
        td_small_MRIO.D_exp,
        nD_exp,
    )
    # test if fp = terr + imp - exp on the total level
    # that tests if imp == exp and fp == terr
    pdt.assert_series_equal(
        nD_cba.sum(axis=1),
        nD_pba.sum(axis=1) + nD_imp.sum(axis=1) - nD_exp.sum(axis=1),
    )



""" Mathematical functions for input output calculations

All methods here should follow the functional programming paradigm

Note
----
To avoid namespace pollution everythin here starts with calc_

"""

import typing
import warnings
from collections import namedtuple

import numpy as np
import pandas as pd

import pymrio.tools.ioutil as ioutil


def calc_x(Z, Y):
    """Calculate the industry output x from the Z and Y matrix

    industry output (x) = flows (sum_columns(Z)) + final demand (sum_columns(Y))

    Parameters
    ----------
    Z : pandas.DataFrame or numpy.array
        Symmetric input output table (flows)
    Y : pandas.DataFrame or numpy.array
        final demand with categories (1.order) for each country (2.order)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Industry output x as column vector
        The type is determined by the type of Z. If DataFrame index as Z

    """
    x = np.reshape(np.sum(np.hstack((Z, Y)), 1), (-1, 1))
    if type(Z) is pd.DataFrame:
        x = pd.DataFrame(x, index=Z.index, columns=["indout"])
    if type(x) is pd.Series:
        x = pd.DataFrame(x)
    if type(x) is pd.DataFrame:
        x.columns = ["indout"]
    return x


def calc_x_from_L(L, y):
    """Calculate the industry output x from L and a y vector

    x = Ly

    The industry output x is computed from a demand vector y

    Parameters
    ----------
    L : pandas.DataFrame or numpy.array
        Symmetric input output Leontief table
    y : pandas.DataFrame or numpy.array
        a column vector of the total final demand

    Returns
    -------
    pandas.DataFrame or numpy.array
        Industry output x as column vector
        The type is determined by the type of L. If DataFrame index as L

    """
    x = L.dot(y)
    if type(x) is pd.Series:
        x = pd.DataFrame(x)
    if type(x) is pd.DataFrame:
        x.columns = ["indout"]
    return x


def calc_Z(A, x):
    """calculate the Z matrix (flows) from A and x

    A = Z / x[None, :]  =>  Z = A * x[None, :]

    By definition, the coefficient matrix A is basically the normalized flows
    So Z is just derived from A by un-normalizing using the industrial output x

    Parameters
    ----------
    A : pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients)
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Symmetric input output table (flows) Z
        The type is determined by the type of A.
        If DataFrame index/columns as A

    """
    if (type(x) is pd.DataFrame) or (type(x) is pd.Series):
        x = x.values
    x = x.reshape((1, -1))  # use numpy broadcasting - much faster
    # (but has to ensure that x is a row vector)
    # old mathematical form:
    # return A.dot(np.diagflat(x))
    if type(A) is pd.DataFrame:
        return pd.DataFrame(A.values * x, index=A.index, columns=A.columns)
    else:
        return A * x


def calc_A(Z, x):
    """Calculate the A matrix (coefficients) from Z and x

    A is a normalized version of the industrial flows Z

    Parameters
    ----------
    Z : pandas.DataFrame or numpy.array
        Symmetric input output table (flows)
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients) A
        The type is determined by the type of Z.
        If DataFrame index/columns as Z

    """
    if (type(x) is pd.DataFrame) or (type(x) is pd.Series):
        x = x.values
    if (type(x) is not np.ndarray) and (x == 0):
        recix = 0
    else:
        with warnings.catch_warnings():
            # catch the divide by zero warning
            # we deal wit that by setting to 0 afterwards
            warnings.simplefilter("ignore")
            recix = 1 / x
        recix[recix == np.inf] = 0
        recix = recix.reshape((1, -1))
    # use numpy broadcasting - factor ten faster
    # Mathematical form - slow
    # return Z.dot(np.diagflat(recix))
    if type(Z) is pd.DataFrame:
        return pd.DataFrame(Z.values * recix, index=Z.index, columns=Z.columns)
    else:
        return Z * recix


def calc_L(A):
    """Calculate the Leontief L from A

    L = inverse matrix of (I - A)

    Where I is an identity matrix of same shape as A

    Comes from:
        x = Ax + y  =>  (I-A)x = y
    Where:
        A: coefficient input () - output () table
        x: output vector
        y: final demand vector

    Hence, L allows to derive a required output vector x for a given demand y

    Parameters
    ----------
    A : pandas.DataFrame or numpy.array
        Symmetric input output table (coefficients)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Leontief input output table L
        The type is determined by the type of A.
        If DataFrame index/columns as A

    """
    I = np.eye(A.shape[0])  # noqa
    if type(A) is pd.DataFrame:
        return pd.DataFrame(np.linalg.inv(I - A), index=A.index, columns=A.columns)
    else:
        return np.linalg.inv(I - A)


def calc_S(F, x):
    """Calculate extensions/factor inputs coefficients

    Parameters
    ----------
    F : pandas.DataFrame or numpy.array
        Total direct impacts
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Direct impact coefficients S
        The type is determined by the type of F.
        If DataFrame index/columns as F

    """
    return calc_A(F, x)


def calc_S_Y(F_Y, y):
    """Calculate extensions/factor inputs coefficients for the final demand

    Note
    ----
    F_Y will be restricted to the item available in y for the calculation. This
    allows to use a subset of Y (just some regions for example) to be used for
    the calculations. Only works when using pandas DataFrames/Series.


    Parameters
    ----------
    F_Y : pandas.DataFrame or numpy.array
         Final demand impacts
    y : pandas.DataFrame or numpy.array
        Final demand column vector (e.g. mrio.Y.sum())

    Returns
    -------
    pandas.DataFrame or numpy.array
        Direct impact coefficients of final demand S_Y
        The type is determined by the type of F.
        If DataFrame index/columns as F

    """
    if type(F_Y) is pd.DataFrame and type(y) is (pd.Series or pd.DataFrame):
        F_Y_calc = F_Y.loc[:, y.index]
    else:
        F_Y_calc = F_Y
    return calc_A(F_Y_calc, y)


def calc_F(S, x):
    """Calculate total direct impacts from the impact coefficients

    Parameters
    ----------
    S : pandas.DataFrame or numpy.array
        Direct impact coefficients S
    x : pandas.DataFrame or numpy.array
        Industry output column vector

    Returns
    -------
    pandas.DataFrame or numpy.array
        Total direct impacts F
        The type is determined by the type of S.
        If DataFrame index/columns as S

    """
    return calc_Z(S, x)


def calc_F_Y(S_Y, y):
    """Calc. total direct impacts from the impact coefficients of final demand

    Note
    ----
    S_Y will be restricted to the item available in y for the calculation. This
    allows to use a subset of Y (just some regions for example) to be used for
    the calculations. Only works when using pandas DataFrames/Series.

    Parameters
    ----------
    S_Y : pandas.DataFrame or numpy.array
        Direct impact coefficients of final demand
    y : pandas.DataFrame or numpy.array
        Final demand column vector (e.g. mrio.Y.sum())

    Returns
    -------
    pandas.DataFrame or numpy.array
        Total direct impacts of final demand F_Y
        The type is determined by the type of S_Y.
        If DataFrame index/columns as S_Y

    """
    if type(S_Y) is pd.DataFrame and type(y) is (pd.Series or pd.DataFrame):
        S_Y_calc = S_Y.loc[:, y.index]
    else:
        S_Y_calc = S_Y

    return calc_Z(S_Y_calc, y)


def calc_M(S, L):
    """Calculate multipliers of the extensions

    Parameters
    ----------
    L : pandas.DataFrame or numpy.array
        Leontief input output table L
    S : pandas.DataFrame or numpy.array
        Direct impact coefficients

    Returns
    -------
    pandas.DataFrame or numpy.array
        Multipliers M
        The type is determined by the type of D.
        If DataFrame index/columns as D

    """
    return S.dot(L)


def calc_e(M, Y):
    """Calculate total impacts (footprints of consumption Y)

    Parameters
    ----------
    M : pandas.DataFrame or numpy.array
        Multipliers
    Y : pandas.DataFrame or numpy.array
        Final consumption

    TODO - this must be completely redone (D, check for dataframe, ...)

    Returns
    -------
    pandas.DataFrame or numpy.array
        Multipliers m
        The type is determined by the type of M.
        If DataFrame index/columns as M
    The calcubased on multipliers M and finald demand Y"""

    return M.dot(Y)


def recalc_M(S, D_cba, Y):
    """Calculate Multipliers based on footprints.

    Parameters
    ----------
    D_cba : pandas.DataFrame or numpy array
        Footprint per sector and country
    Y : pandas.DataFrame or numpy array
        Final demand: aggregated across categories or just one category, one
        column per country. This will be diagonalized per country block.
        The diagonolized form must be invertable for this method to work.

    Returns
    -------

    pandas.DataFrame or numpy.array
        Multipliers M
        The type is determined by the type of D_cba.
        If DataFrame index/columns as D_cba


    """

    Y_diag = ioutil.diagonalize_columns_to_sectors(Y)
    Y_inv = np.linalg.inv(Y_diag)
    M = D_cba.dot(Y_inv)
    if type(D_cba) is pd.DataFrame:
        M.columns = D_cba.columns
        M.index = D_cba.index

    return M


def calc_accounts(S, L, Y):
    """Calculate sector specific cba and pba based accounts, imp and exp accounts

    The total industry output x for the calculation
    is recalculated from L and y

    Parameters
    ----------
    L : pandas.DataFrame
        Leontief input output table L
    S : pandas.DataFrame
        Direct impact coefficients
    Y : pandas.DataFrame
        Final demand: aggregated across categories or just one category, one
        column per country. Shape: rows as L, Y with just region names

    Returns
    -------
    Tuple
        (D_cba, D_pba, D_imp, D_exp)

        Format: D_row x L_col (=nr_countries*nr_sectors)

        - D_cba        Footprint per sector and country
        - D_pba      Total factur use per sector and country
        - D_imp       Total global factor use to satisfy total final demand in
                      the country per sector
        - D_exp       Total factor use in one country to satisfy final demand
                      in all other countries (per sector)
    """
    # diagonalize each sector block per country
    # this results in a disaggregated y with final demand per country per
    # sector in one column

    if isinstance(Y.columns, pd.MultiIndex):
        raise ValueError(
            "Column index of Y can not be a MultiIndex - aggregate the columns"
        )
    Y_diag = ioutil.diagonalize_columns_to_sectors(Y)
    x_diag = L @ Y_diag

    x_tot = x_diag.values.sum(1)

    del Y_diag

    D_cba = pd.DataFrame(S @ x_diag)

    # D_pba = S.dot(np.diagflat(x_tot))
    # faster broadcasted calculation:
    # NB: D_pba columns might be different to D_cba columns if Y include "regions" which are not in the core. This happens for example for statistical discrepancy in the OECD tables. It is "theoretically" possible to calculate footprints for these "regions", but not PBA accounts.
    D_pba = pd.DataFrame(
        S.values * x_tot.reshape((1, -1)), index=S.index, columns=S.columns
    )

    # for the traded accounts set the domestic industry output to zero
    x_trade = ioutil.set_dom_block(x_diag, value=0)
    D_imp = pd.DataFrame(S @ x_trade)

    x_exp = x_trade.sum(1).values

    # D_exp = S.dot(np.diagflat(x_exp))
    # faster broadcasted version:
    D_exp = pd.DataFrame(
        S.values * x_exp.reshape((1, -1)), index=S.index, columns=S.columns
    )

    return (D_cba, D_pba, D_imp, D_exp)


def calc_gross_trade(
    Z: pd.DataFrame, Y: pd.DataFrame
) -> typing.NamedTuple(
    "gross_trade", [("bilat_flows", pd.DataFrame), ("totals", pd.DataFrame)]
):
    """Calculate the gross bilateral trade flows and totals

    These are the entries of Z and Y with the domestic blocks set to 0.

    Notes
    ----------
    This only works for DataFrame representation of Z and Y following the
    standard pymrio structure (regions on Multiindex level 0 or named 'region',
    sectors/categories on Multiindex level 1 or named 'sector').

    Parameters
    ----------
    Z : pandas.DataFrame
        Symmetric input output table (flows)
    Y : pandas.DataFrame
        final demand with regions (multiindex level 1) and categories (level 2)

    Returns
    -------
    namedtuple (with two DataFrames)
        A NamedTuple with two fields:

            - bilat_flows: df with rows: exporting country and sector,
              columns: importing countries
            - totals: df with gross total imports and exports per sector
              and region


    """

    Z_trade_blocks = ioutil.set_dom_block(Z, value=0)
    Y_trade_blocks = ioutil.set_dom_block(Y, value=0)

    level_spec_Z = "region" if "region" in Z.columns.names else 0
    level_spec_Y = "region" if "region" in Y.columns.names else 0
    Z_trade_agg = Z_trade_blocks.groupby(axis=1, level=level_spec_Z, sort=False).agg(
        sum
    )
    Y_trade_agg = Y_trade_blocks.groupby(axis=1, level=level_spec_Y, sort=False).agg(
        sum
    )

    x_bilat = Z_trade_agg + Y_trade_agg

    level_spec_x = "sector" if "sector" in x_bilat.index.names else 1
    gross_imports = pd.DataFrame(
        x_bilat.groupby(axis=0, level=level_spec_x, sort=False)
        .agg(sum)
        .stack()
        .swaplevel(),
        columns=["imports"],
    )
    gross_exports = pd.DataFrame(x_bilat.sum(axis=1), columns=["exports"])

    # gross_exports first as the swaplevel in gross_imports rearanges sectors!
    gross_totals = gross_exports.join(gross_imports)

    return namedtuple("gross_trade", "bilat_flows totals")(x_bilat, gross_totals)



""" Testing functions for the full run based on
the small MRIO given within pymrio.

This tests the full computation and fileio.

Might be the slowest test to run - make optional if it takes to long.
"""

import os
import sys

import numpy as np
import numpy.testing as npt
import pytest

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

import pymrio  # noqa


@pytest.fixture()
def td_testmrio():
    """Results from the test mrio calculation.

    This works with the test system given in pymrio.

    """

    class results:
        class factor_inputs:
            D_imp_values = np.array(
                [
                    [
                        2382767.5387909594,
                        2393134.584113787,
                        2972317.803194902,
                        4201655.226133792,
                        1428782.8385865367,
                        6693276.934870526,
                    ]  # noqa
                ]
            )

    return results


def test_all(td_testmrio):
    """Full integration test

    Checks:
    1) the cba calculation
    2) concate extension
    """
    mr = pymrio.load_test()
    mr.calc_all()
    npt.assert_allclose(
        td_testmrio.factor_inputs.D_imp_values,
        mr.factor_inputs.D_imp_reg.values,
        rtol=1e-5,
    )
    sat_new = pymrio.concate_extension(mr.emissions, mr.factor_inputs, name="sat_new")

    assert len(sat_new.D_cba) == 3
    assert mr.emissions.F.index[0] in sat_new.F.index
    assert mr.factor_inputs.F.index[0] in sat_new.F_Y.index
    assert all(mr.emissions.get_regions() == sat_new.get_regions())
    assert all(mr.emissions.get_sectors() == sat_new.get_sectors())


def test_txt_zip_fileio(tmpdir):
    """Round trip with saving and loading the testmrio

    Txt and zip format

    Also tests some fileio related util functions
    """
    mr = pymrio.load_test()

    save_path = str(tmpdir.mkdir("pymrio_test"))

    with pytest.raises(ValueError):
        mr.save_all(save_path, table_format="nan")

    mr.save_all(save_path)

    # Testing getting the repo content, some random sample test
    fc = pymrio.get_repo_content(save_path)
    assert fc.iszip is False
    assert "Z.txt" in [os.path.basename(f) for f in fc.filelist]
    assert "F_Y.txt" in [os.path.basename(f) for f in fc.filelist]
    assert "unit.txt" in [os.path.basename(f) for f in fc.filelist]

    mr2 = pymrio.load_all(save_path)
    npt.assert_allclose(mr.Z.values, mr2.Z.values, rtol=1e-5)

    # Testing the zip archive functions
    zip_arc = os.path.join(save_path, "test_mrio.zip")
    if os.path.exists(zip_arc):
        os.remove(zip_arc)

    with pytest.raises(FileExistsError):
        pymrio.archive(
            source=fc.filelist,
            archive=zip_arc,
            remove_source=False,
            path_in_arc="test_duplicates",
        )

        pymrio.archive(
            source=fc.filelist,
            archive=zip_arc,
            remove_source=False,
            path_in_arc="test_duplicates",
        )

    if os.path.exists(zip_arc):
        os.remove(zip_arc)

    pymrio.archive(
        source=save_path, archive=zip_arc, remove_source=True, path_in_arc="test"
    )

    fc = pymrio.get_repo_content(zip_arc)
    assert fc.iszip is True
    assert "Z.txt" in [os.path.basename(f) for f in fc.filelist]
    assert "F_Y.txt" in [os.path.basename(f) for f in fc.filelist]
    assert "unit.txt" in [os.path.basename(f) for f in fc.filelist]

    emissions = pymrio.load(zip_arc, path_in_arc="test/emissions")
    npt.assert_allclose(mr.emissions.F.values, emissions.F.values, rtol=1e-5)

    mr3 = pymrio.load_all(zip_arc)
    npt.assert_allclose(mr.Z.values, mr3.Z.values, rtol=1e-5)

    with pytest.raises(pymrio.ReadError):
        pymrio.load_all(zip_arc, path_in_arc="./foo")

    with pytest.raises(pymrio.ReadError):
        pymrio.load(path="./foo")


def test_parquet_fileio(tmpdir):
    """Round trip with saving and loading the testmrio in parquet"""
    mr = pymrio.load_test()

    save_path = str(tmpdir.mkdir("pymrio_test_parquet"))

    mr.save_all(save_path, table_format="parquet")

    mr_load = pymrio.load_all(save_path)

    assert mr_load == mr


def test_pickle_fileio(tmpdir):
    """Round trip with saving and loading the testmrio in pickle"""
    mr = pymrio.load_test()

    save_path = str(tmpdir.mkdir("pymrio_test_pickle"))

    mr.save_all(save_path, table_format="pickle")

    mr_load = pymrio.load_all(save_path)

    assert mr_load == mr


def test_reports(tmpdir):
    """Tests the reporting function

    Here, only the correct file and folder structure
    is tested. The actual graphical output gets tested
    in test_outputs.
    """
    mr = pymrio.load_test().calc_all()

    save_path = str(tmpdir.mkdir("pymrio_test_reports"))

    with pytest.raises(ValueError):
        mr.report_accounts(
            path=save_path, per_capita=False, per_region=False, format="html"
        )

    mr.report_accounts(path=save_path, per_capita=True, format="html")

    file_content = os.listdir(save_path)
    for ext in mr.get_extensions(data=True):
        assert ext.name.replace(" ", "_") + "_per_capita.html" in file_content
        assert ext.name.replace(" ", "_") + "_per_region.html" in file_content

        cap_pic_file_content = os.listdir(
            os.path.join(save_path, ext.name.replace(" ", "_") + "_per_capita")
        )
        reg_pic_file_content = os.listdir(
            os.path.join(save_path, ext.name.replace(" ", "_") + "_per_region")
        )

        assert "png" in cap_pic_file_content[0]
        assert "png" in reg_pic_file_content[0]



""" Utility functions for automatic downloading of public MRIO databases
"""

import getpass
import itertools
import json
import os
import re
import ssl
import zipfile
from collections import namedtuple

import requests
import urllib3

from pymrio.core.constants import __ROOT, GLORIA_URLS
from pymrio.tools.iometadata import MRIOMetaData
from pymrio.tools.ioutil import filename_from_url, ssl_fix

WIOD_CONFIG = {
    "url_db_view": "http://www.wiod.org/database/wiots13",
    "url_db_content": "http://www.wiod.org/",
    "mrio_regex": r"protected.*?wiot\d\d.*?xlsx",
    "satellite_urls": [
        "http://www.wiod.org/protected3/data13/SEA/WIOD_SEA_July14.xlsx",
        "http://www.wiod.org/protected3/data13/EU/EU_may12.zip",
        "http://www.wiod.org/protected3/data13/EM/EM_may12.zip",
        "http://www.wiod.org/protected3/data13/CO2/CO2_may12.zip",
        "http://www.wiod.org/protected3/data13/AIR/AIR_may12.zip",
        "http://www.wiod.org/protected3/data13/land/lan_may12.zip",
        "http://www.wiod.org/protected3/data13/materials/mat_may12.zip",
        "http://www.wiod.org/protected3/data13/water/wat_may12.zip",
    ],
}

EORA26_CONFIG = {
    "url_db_view": "http://worldmrio.com/simplified/",
    "url_db_content": "http://worldmrio.com/",
}

EXIOBASE3_CONFIG = {
    "url_db_view": "https://doi.org/10.5281/zenodo.3583070",  # lastest version
    # "url_db_view": "https://doi.org/10.5281/zenodo.3583071",  # version 3.7
    # "url_db_view": "https://doi.org/10.5281/zenodo.4277368",  # version 3.8
    "url_db_content": "",
    "mrio_regex": r"https://zenodo.org/record/\d*/files/IOT_\d\d\d\d_[p,i]x[p,i].zip",
    "requests_func": requests.get,
}


OECD_CONFIG = {
    "url_db_view": "https://www.oecd.org/sti/ind/inter-country-input-output-tables.htm",  # NOQA
    "url_db_content": "https://www.oecd.org/sti/ind/",
    "mrio_regex": r"ICIO\d\d\d\d_\d\d\d\d\.zip",
    "datafiles": {
        "v2016": {
            "1995": "https://www.oecd.org/sti/ind/ICIO2016_1995.zip",
            "1996": "https://www.oecd.org/sti/ind/ICIO2016_1996.zip",
            "1997": "https://www.oecd.org/sti/ind/ICIO2016_1997.zip",
            "1998": "https://www.oecd.org/sti/ind/ICIO2016_1998.zip",
            "1999": "https://www.oecd.org/sti/ind/ICIO2016_1999.zip",
            "2000": "https://www.oecd.org/sti/ind/ICIO2016_2000.zip",
            "2001": "https://www.oecd.org/sti/ind/ICIO2016_2001.zip",
            "2002": "https://www.oecd.org/sti/ind/ICIO2016_2002.zip",
            "2003": "https://www.oecd.org/sti/ind/ICIO2016_2003.zip",
            "2004": "https://www.oecd.org/sti/ind/ICIO2016_2004.zip",
            "2005": "https://www.oecd.org/sti/ind/ICIO2016_2005.zip",
            "2006": "https://www.oecd.org/sti/ind/ICIO2016_2006.zip",
            "2007": "https://www.oecd.org/sti/ind/ICIO2016_2007.zip",
            "2008": "https://www.oecd.org/sti/ind/ICIO2016_2008.zip",
            "2009": "https://www.oecd.org/sti/ind/ICIO2016_2009.zip",
            "2010": "https://www.oecd.org/sti/ind/ICIO2016_2010.zip",
            "2011": "https://www.oecd.org/sti/ind/ICIO2016_2011.zip",
        },
        "v2018": {
            "2005": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=1f134869-1820-49ce-b8b8-3973ec8db607",  # NOQA
            "2006": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=da62c835-f4fa-4450-bf19-1dd60f88a385",  # NOQA
            "2007": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=c4d4c21d-00db-48d8-9f9a-f722fcdca494",  # NOQA
            "2008": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=1fd2fc03-c140-46f4-818e-9a66b671ff70",  # NOQA
            "2009": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=4cc79090-d1ee-48b6-a252-e75312d32a1c",  # NOQA
            "2010": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=16d04830-3c27-47a5-bc03-e429d27f585e",  # NOQA
            "2011": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=dc48c8c0-f200-487a-aecb-0c2c17fe3ddf",  # NOQA
            "2012": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=cfd03495-8a90-4449-8097-a30f06853cab",  # NOQA
            "2013": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=8c8ac674-1b6c-4c8e-94d1-158f06285659",  # NOQA
            "2014": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=0190bd9d-31d0-4171-bd1c-82d96b88e469",  # NOQA
            "2015": "http://stats.oecd.org/wbos/fileview2.aspx?IDFile=9f579ef3-4685-45e4-a0ba-d1acbd9755a6",  # NOQA
        },
        "v2021": {
            "1995-1999": "https://stats.oecd.org/wbos/fileview2.aspx?IDFile=91d8e84b-7406-46b9-af5f-ec096242755c",
            "2000-2004": "https://stats.oecd.org/wbos/fileview2.aspx?IDFile=8adf89dd-18b4-40fe-bc7f-c822052eb961",
            "2005-2009": "https://stats.oecd.org/wbos/fileview2.aspx?IDFile=fe218690-0a3b-44aa-a82c-b3e3da6d24db",
            "2010-2014": "https://stats.oecd.org/wbos/fileview2.aspx?IDFile=2c2f499f-5703-4034-9457-2f7518e8f2fc",
            "2015-2018": "https://stats.oecd.org/wbos/fileview2.aspx?IDFile=59a3d7f2-3f23-40d5-95ca-48da84c0f861",
        },
    },
}

GLORIA_CONFIG = {"datafiles": GLORIA_URLS}


def _get_url_datafiles(
    url_db_view,
    url_db_content,
    mrio_regex,
    access_cookie=None,
    requests_func=requests.post,
):
    """Urls of mrio files by parsing url content for mrio_regex

    Parameters
    ----------

    url_db_view: url str
        Url which shows the list of mrios in the db

    url_db_content: url str
        Url which needs to be appended before the url parsed from the
        url_db_view to get a valid download link

    mrio_regex: regex str
        Regex to parse the mrio datafile from url_db_view

    access_cookie: dict, optional
        If needed, cookie to access the database

    requests_func: function
        Function to use for retrieving the url content.
        Can be requests.get or requests.post

    Returns
    -------
    Named tuple:
    .raw_text: content of url_db_view for later use
    .data_urls: list of url

    """
    # Use post here - NB: get could be necessary for some other pages
    # but currently works for wiod and eora
    returnvalue = namedtuple("url_content", ["raw_text", "data_urls"])
    url_text = requests_func(url_db_view, cookies=access_cookie).text
    data_urls = [url_db_content + ff for ff in re.findall(mrio_regex, url_text)]
    return returnvalue(raw_text=url_text, data_urls=data_urls)


def _download_urls(
    url_list, storage_folder, overwrite_existing, downlog_handler, access_cookie=None
):
    """Save url from url_list to storage_folder

    Parameters
    ----------
    url_list: list of str
        Valid url to download

    storage_folder: str, valid path
        Location to store the download, folder will be created if
        not existing. If the file is already present in the folder,
        the download depends on the setting in 'overwrite_existing'.

    overwrite_existing: boolean, optional
        If False, skip download of file already existing in
        the storage folder (default). Set to True to replace
        files.

    downlog_handler: instance of MRIOMetaData
        Instance of MRIOMetaData to store the download log

    access_cookie: cookie, optional
        Cookie to be passed to the requests.post function fetching the data


    Returns
    -------

    The downlog_handler is passed back

    """
    for url in url_list:
        filename = filename_from_url(url)
        if downlog_handler.name == "Eora":
            filename = filename.split(".zip")[0] + ".zip"
        if not overwrite_existing and filename in os.listdir(storage_folder):
            continue
        storage_file = os.path.join(storage_folder, filename)

        # Using requests here - tried with aiohttp but was actually slower
        # Also don’t use shutil.copyfileobj - corrupts zips from Eora
        # req = requests.post(url, stream=True, cookies=access_cookie)
        req = requests.get(url, stream=True, cookies=access_cookie)
        with open(storage_file, "wb") as lf:
            for chunk in req.iter_content(1024 * 5):
                lf.write(chunk)

        downlog_handler._add_fileio("Downloaded {} to {}".format(url, filename))
        downlog_handler.save()

    return downlog_handler


def download_oecd(
    storage_folder, version="v2021", years=None, overwrite_existing=False
):
    """Downloads the OECD ICIO tables

    Parameters
    ----------
    storage_folder: str, valid path
        Location to store the download, folder will be created if
        not existing. If the file is already present in the folder,
        the download of the specific file will be skipped.

    version: string or int, optional
        Two versions of the ICIO OECD tables are currently availabe:
        Version >v2016<: based on >SNA93< / >ISIC Rev.3<
        Version >v2018<: based on >SNA08< / >ISIC Rev.4< (default)
        Pass any of the identifiers between >< to specifiy the
        version to be downloaded.

    years: list of int (4 digit) or str, optional
        If years is given only downloads the specific years.

    overwrite_existing: boolean, optional
        If False, skip download of file already existing in
        the storage folder (default). Set to True to replace
        files.

    Returns
    -------

    Meta data of the downloaded MRIOs

    """
    # Implementation Notes:
    # For OECD the generic download routines can not be used
    # b/c the 2018 version is coded as aspx fileview property
    # in the html source - instead a hardcoded dict is used
    # to select the url for download

    os.makedirs(storage_folder, exist_ok=True)

    if type(version) is int:
        version = str(version)

    if ("8" in version) or ("4" in version):
        version = "v2018"
    elif ("3" in version) or ("6" in version):
        version = "v2016"
    elif "21" in version:
        version = "v2021"
    else:
        raise ValueError("Version not understood")

    if type(years) is int or type(years) is str:
        years = [years]

    if not years:
        if version == "v2018":
            years = range(2005, 2016)
        elif version == "v2021":
            years = ["1995-1999", "2000-2004", "2005-2009", "2010-2014", "2015-2018"]

        else:
            years = range(1995, 2012)

    years = [str(yy) for yy in years]

    downlog = MRIOMetaData._make_download_log(
        location=storage_folder,
        description="OECD-ICIO download",
        name="OECD-ICIO",
        system="IxI",
        version=version,
    )

    for yy in years:
        if yy not in OECD_CONFIG["datafiles"][version].keys():
            raise ValueError("Datafile for {} not specified or available.".format(yy))

        filename = "ICIO" + version.lstrip("v") + "_" + yy + ".zip"

        if not overwrite_existing:
            if version == "v2021":
                filenames = [
                    f"ICIO{version.lstrip('v')}_{str(yr)}.csv"
                    for yr in range(int(yy[:4]), int(yy[-4:]) + 1)
                ]
                if set(filenames).issubset(os.listdir(storage_folder)):
                    continue
            elif filename in os.listdir(storage_folder):
                continue

        req = ssl_fix(OECD_CONFIG["datafiles"][version][yy], stream=True)
        storage_file = os.path.join(storage_folder, filename)

        with open(storage_file, "wb") as lf:
            for chunk in req.iter_content(1024 * 5):
                lf.write(chunk)

        if version == "v2021":
            with zipfile.ZipFile(storage_file, "r") as zip_ref:
                zip_ref.extractall(storage_folder)
            os.remove(storage_file)

        downlog._add_fileio(
            "Downloaded {} to {}".format(
                OECD_CONFIG["datafiles"][version][yy], filename
            )
        )

    downlog.save()

    return downlog


def download_wiod2013(
    storage_folder,
    years=None,
    overwrite_existing=False,
    satellite_urls=WIOD_CONFIG["satellite_urls"],
):
    """Downloads the 2013 wiod release

    Note
    ----
    Currently, pymrio only works with the 2013 release of the wiod tables. The
    more recent 2016 release so far (October 2017) lacks the environmental and
    social extensions.


    Parameters
    ----------
    storage_folder: str, valid path
        Location to store the download, folder will be created if
        not existing. If the file is already present in the folder,
        the download of the specific file will be skipped.


    years: list of int or str, optional
        If years is given only downloads the specific years. This
        only applies to the IO tables because extensions are stored
        by country and not per year.
        The years can be given in 2 or 4 digits.

    overwrite_existing: boolean, optional
        If False, skip download of file already existing in
        the storage folder (default). Set to True to replace
        files.

    satellite_urls : list of str (urls), optional
        Which satellite accounts to download.  Default: satellite urls defined
        in WIOD_CONFIG - list of all available urls Remove items from this list
        to only download a subset of extensions

    Returns
    -------

    Meta data of the downloaded MRIOs

    """

    os.makedirs(storage_folder, exist_ok=True)

    if type(years) is int or type(years) is str:
        years = [years]
    years = years if years else range(1995, 2012)
    years = [str(yy).zfill(2)[-2:] for yy in years]

    wiod_web_content = _get_url_datafiles(
        url_db_view=WIOD_CONFIG["url_db_view"],
        url_db_content=WIOD_CONFIG["url_db_content"],
        mrio_regex=WIOD_CONFIG["mrio_regex"],
    )

    restricted_wiod_io_urls = [
        url
        for url in wiod_web_content.data_urls
        if re.search(r"(wiot)(\d\d)", os.path.basename(url)).group(2) in years
    ]

    downlog = MRIOMetaData._make_download_log(
        location=storage_folder,
        description="WIOD metadata file for pymrio",
        name="WIOD",
        system="IxI",
        version="data13",
    )

    downlog = _download_urls(
        url_list=restricted_wiod_io_urls + satellite_urls,
        storage_folder=storage_folder,
        overwrite_existing=overwrite_existing,
        downlog_handler=downlog,
    )

    downlog.save()
    return downlog


def download_eora26(
    storage_folder, email, password, years=None, prices=["bp"], overwrite_existing=False
):
    """Downloading eora26 mrios (registration required),
    To use this function you have to have an Eora account,
    New account registration can be done through https://worldmrio.com/login.jsp

    Parameters
    ----------
    storage_folder: str, valid path
        Location to store the download, folder will be created if
        not existing. If the file is already present in the folder,
        the download of the specific file will be skipped.
    email: str,
        Eora account email
    password: str,
        Eora account password
    years: list of int or str, optional
        If years is given only downloads the specific years. This
        only applies to the IO tables because extensions are stored
        by country and not per year.
        The years can be given in 2 or 4 digits.
    prices: list of str
        If bp (default), download basic price tables.
        If pp, download purchaser prices. ['bp', 'pp'] possible.
    overwrite_existing: boolean, optional
        If False, skip download of file already existing in
        the storage folder (default). Set to True to replace
        files.
    """

    try:
        os.makedirs(storage_folder)
    except FileExistsError:
        pass

    false_cred = True

    while false_cred:
        r = requests.post(
            "https://worldmrio.com/Register2?submit=login",
            data={
                "email": email,
                "pass": password,
                "targetURL": "null",
                "submit": "login",
            },
        )

        if "no account found" in r.text:
            print(
                """Eora account with this email was not found\n
            Please try again or register a new user at the following website:\n
            https://worldmrio.com/login.jsp"""
            )

            email = input("Enter your Eora account email: ")
            password = getpass.getpass(prompt="Enter your Eora account password: ")

        elif "Sorry, wrong password provided" in r.text:
            print(
                """The password for this email account is incorrect\n
            Please try again"""
            )

            password = getpass.getpass(prompt="Enter your Eora account password: ")

        else:
            false_cred = False

    if type(years) is int or type(years) is str:
        years = [years]
    years = years if years else range(1995, 2012)
    years = [str(yy).zfill(4) for yy in years]

    if type(prices) is str:
        prices = [prices]

    restricted_eora_urls = [
        f"https://worldmrio.com/ComputationsM/Phase199/Loop082/simplified/Eora26_{yr}_bp.zip?email={email}&pass={password}"
        for yr in years
    ]

    downlog = MRIOMetaData._make_download_log(
        location=storage_folder,
        description="Download log for Eora",
        name="Eora",
        system="ixi",
        version="v199.82",
    )

    downlog = _download_urls(
        url_list=restricted_eora_urls,
        storage_folder=storage_folder,
        overwrite_existing=overwrite_existing,
        downlog_handler=downlog,
    )

    downlog.save()

    return downlog


def download_exiobase1():
    """Downloading exiobase not implemented (registration required)"""
    raise NotImplementedError(
        "EXIOBASE 1 requires registration prior to download. "
        "Please register at www.exiobase.eu and download the "
        "EXIOBASE MRIO files "
        "pxp_ita_44_regions_coeff_txt or "
        "ixi_fpa_44_regions_coeff_txt "
        "manually (tab Data Download - EXIOBASE 1 (full data set)."
    )
    return None


def download_exiobase2():
    """Downloading exiobase not implemented (registration required)"""
    raise NotImplementedError(
        "EXIOBASE 2 requires registration prior to download. "
        "Please register at www.exiobase.eu and download the "
        "EXIOBASE MRIO files "
        ">MrIOT_IxI_fpa_coefficient_version2.2.2.zip<_"
        "and/or_"
        ">MrIOT_PxP_ita_coefficient_version2.2.2.zip<_"
        "manually (tab Data Download - EXIOBASE 2)."
    )
    return None


def download_exiobase3(
    storage_folder,
    years=None,
    system=None,
    overwrite_existing=False,
    doi="10.5281/zenodo.3583070",
):
    """
    Downloads EXIOBASE 3 files from Zenodo

    Since version 3.7 EXIOBASE gets published on the Zenodo scientific data
    repository.  This function download the lastest available version from
    Zenodo, for previous version the corresponding DOI (parameter 'doi') needs
    to specified.

    Version 3.7: 10.5281/zenodo.3583071
    Version 3.8: 10.5281/zenodo.4277368


    Parameters
    ----------
    storage_folder: str, valid path
        Location to store the download, folder will be created if
        not existing. If the file is already present in the folder,
        the download of the specific file will be skipped.


    years: list of int or str, optional
        If years is given only downloads the specific years (be default all years will be downloaded).
        Years must be given in 4 digits.

    system: string or list of strings, optional
        'pxp': download product by product classification
        'ixi': download industry by industry classification
        ['ixi', 'pxp'] or None (default): download both classifications

    overwrite_existing: boolean, optional
        If False, skip download of file already existing in
        the storage folder (default). Set to True to replace
        files.

    doi: string, optional.
        The EXIOBASE DOI to be downloaded. By default that resolves
        to the DOI citing the latest available version. For the previous DOI
        see the block 'Versions' on the right hand side of
        https://zenodo.org/record/4277368.

    Returns
    -------

    Meta data of the downloaded MRIOs

    """

    os.makedirs(storage_folder, exist_ok=True)

    doi_url = "https://doi.org/" + doi
    EXIOBASE3_CONFIG["url_db_view"] = doi_url

    exio_web_content = _get_url_datafiles(**EXIOBASE3_CONFIG)

    file_pattern = re.compile(r"IOT_[1,2]\d\d\d_[p,i]x[p,i]\.zip")
    available_files = [
        file_pattern.search(url).group() for url in exio_web_content.data_urls
    ]

    available_years = {filename.split("_")[1] for filename in available_files}
    if type(years) is int or type(years) is str:
        years = [years]
    years = years if years else list(available_years)

    system = system if system else ["pxp", "ixi"]
    if type(system) is str:
        system = [system]

    downlog = MRIOMetaData._make_download_log(
        location=storage_folder,
        description="Download log of EXIOBASE3",
        name="EXIO3",
        system=",".join(system),
        version=doi,
    )

    requested_urls = []
    for file_specs in itertools.product(years, system):
        filename = list(
            filter(
                lambda x: str(file_specs[0]) in x and str(file_specs[1]) in x,
                available_files,
            )
        )

        if not filename:
            downlog._add_fileio(
                "Could not find EXIOBASE 3 source file with >{}< and >{}<".format(
                    file_specs[0], file_specs[1]
                )
            )
            continue
        requested_urls += [
            u for u in exio_web_content.data_urls for f in filename if f in u
        ]

    downlog = _download_urls(
        url_list=requested_urls,
        storage_folder=storage_folder,
        overwrite_existing=overwrite_existing,
        downlog_handler=downlog,
    )

    downlog.save()
    return downlog


def download_gloria(
    storage_folder,
    urls=GLORIA_CONFIG["datafiles"],
    year=None,
    version=57,
    overwrite_existing=False,
):
    """
    Download Gloria databases files

    Parameters
    ----------

    urls: dict, optional
        Dictionary containing the links of gloria databases
        for different versions, this is already fed to the function,
        imported from urls.json file

    storage_folder: str, option
        The path where to download the file(s), if not specified
        it/they will be downloaded to the current working directory

    year: int, str or list, optional
        The year(s) of the wanted database, if not specified
        the databases of all available years will be downloaded

    version: int or str, option
        The wanted version of Gloria database, if not specified
        the database of the latest version will be downloaded

    overwrite_existing: boolean, optional
        If False, skip download of file already existing in
        the storage folder (default). Set to True to replace
        files.

    Returns
    -------

    No returns
    """

    if f"0{int(version)}" not in urls.keys():
        raise Exception("Specified version is invalid")

    downlog = MRIOMetaData._make_download_log(
        location=storage_folder,
        description="Download log of Gloria",
        name="GLORIA",
        system="IxI",
        version=version,
    )

    files_to_download = []
    if type(year) is int or type(year) is str:
        year = [year]

    if year:
        for yr in year:
            files_to_download.extend(
                [
                    file
                    for file in urls[f"0{int(version)}"]
                    if str(yr) in filename_from_url(file)
                ]
            )
    else:
        files_to_download = urls[f"0{int(version)}"]

    downlog = _download_urls(
        url_list=files_to_download,
        storage_folder=storage_folder,
        overwrite_existing=overwrite_existing,
        downlog_handler=downlog,
    )

    downlog.save()
    return downlog



"""
Utility function for pymrio

KST 20140502
"""
import json
import logging
import os
import re
import ssl
import zipfile
from collections import namedtuple
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import requests
import urllib3

from pymrio.core.constants import DEFAULT_FILE_NAMES, PYMRIO_PATH


def is_vector(inp):
    """Returns true if the input can be interpreted as a 'true' vector

    Note
    ----
    Does only check dimensions, not if type is numeric

    Parameters
    ----------
    inp : numpy.ndarray or something that can be converted into ndarray

    Returns
    -------
    Boolean
        True for vectors: ndim = 1 or ndim = 2 and shape of one axis = 1
        False for all other arrays
    """
    inp = np.asarray(inp)
    nr_dim = np.ndim(inp)
    if nr_dim == 1:
        return True
    elif (nr_dim == 2) and (1 in inp.shape):
        return True
    else:
        return False


def get_repo_content(path):
    """List of files in a repo (path or zip)


    Parameters
    ----------

    path: string or pathlib.Path

    Returns
    -------

    Returns a namedtuple with .iszip and .filelist
    The path in filelist are pure strings.

    """
    path = Path(path)

    if zipfile.is_zipfile(str(path)):
        with zipfile.ZipFile(str(path)) as zz:
            filelist = [info.filename for info in zz.infolist()]
        iszip = True
    else:
        iszip = False
        filelist = [str(f) for f in path.glob("**/*") if f.is_file()]
    return namedtuple("repocontent", ["iszip", "filelist"])(iszip, filelist)


def get_file_para(path, path_in_arc=""):
    """Generic method to read the file parameter file

    Helper function to consistently read the file parameter file, which can
    either be uncompressed or included in a zip archive.  By default, the file
    name is to be expected as set in DEFAULT_FILE_NAMES['filepara'] (currently
    file_parameters.json), but can defined otherwise by including the file
    name of the parameter file in the parameter path.

    Parameters
    ----------

    path: pathlib.Path or string
        Path or path with para file name for the data to load.
        This must either point to the directory containing the uncompressed
        data or the location of a compressed zip file with the data. In the
        later case the parameter 'path_in_arc' needs to be specific to
        further indicate the location of the data in the compressed file.

    path_in_arc: string, optional
        Path to the data in the zip file (where the fileparameters file is
        located). path_in_arc must be given without leading dot and slash;
        thus to point to the data in the root of the compressed file pass ''
        (default), for data in e.g. the folder 'emissions' pass 'emissions/'.
        Only used if parameter 'path' points to an compressed zip file.

    Returns
    -------

    Returns a namedtuple with
    .folder: str with the absolute path containing the
           file parameter file. In case of a zip the path
           is relative to the root in the zip
    .name: Filename without folder of the used parameter file.
    .content: Dictionary with the content oft the file parameter file

    Raises
    ------

    FileNotFoundError if parameter file not found

    """
    path = Path(path)

    if zipfile.is_zipfile(str(path)):
        para_file_folder = str(path_in_arc)
        with zipfile.ZipFile(file=str(path)) as zf:
            files = zf.namelist()
    else:
        para_file_folder = str(path)
        files = [str(f) for f in path.glob("**/*")]

    if para_file_folder not in files:
        if zipfile.is_zipfile(str(path)):
            # b/c in win os.path.join adds \ also for within zipfile
            if para_file_folder != "":
                para_file_full_path = (
                    para_file_folder + "/" + DEFAULT_FILE_NAMES["filepara"]
                ).replace("//", "/")
            else:
                para_file_full_path = DEFAULT_FILE_NAMES["filepara"]

        else:
            para_file_full_path = os.path.join(
                para_file_folder, DEFAULT_FILE_NAMES["filepara"]
            )
    else:
        para_file_full_path = para_file_folder
        para_file_folder = os.path.dirname(para_file_full_path)

    if para_file_full_path not in files:
        raise FileNotFoundError(
            "File parameter file {} not found".format(para_file_full_path)
        )

    if zipfile.is_zipfile(str(path)):
        with zipfile.ZipFile(file=str(path)) as zf:
            para_file_content = json.loads(zf.read(para_file_full_path).decode("utf-8"))
    else:
        with open(para_file_full_path, "r") as pf:
            para_file_content = json.load(pf)

    return namedtuple("file_parameter", ["folder", "name", "content"])(
        para_file_folder, os.path.basename(para_file_full_path), para_file_content
    )


def build_agg_matrix(agg_vector, pos_dict=None):
    """Agg. matrix based on mapping given in input as numerical or str vector.

    The aggregation matrix has the from nxm with

    -n  new classificaction
    -m  old classification

    Parameters
    ----------
        agg_vector : list or vector like numpy ndarray
            This can be row or column vector.
            Length m with position given for n and -1 if values
            should not be included
            or
            length m with id_string for the aggregation

        pos_dict : dictionary
            (only possible if agg_vector is given as string)
            output order for the new matrix
            must be given as dict with
            'string in agg_vector' = pos
            (as int, -1 if value should not be included in the aggregation)

    Example 1:
        input vector: np.array([0, 1, 1, 2]) or ['a', 'b', 'b', 'c']

        agg matrix:

           m0  m1  m2  m3
        n0 1   0   0   0
        n1 0   1   1   0
        n2 0   0   0   1

    Example 2:
        input vector: np.array([1, 0, 0, 2]) or
                      (['b', 'a', 'a', 'c'], dict(a=0,b=1,c=2))

        agg matrix:

           m0  m1  m2  m3
        n0 0   1   1   0
        n1 1   0   0   0
        n2 0   0   0   1
    """
    if isinstance(agg_vector, np.ndarray):
        agg_vector = agg_vector.flatten().tolist()

    if type(agg_vector[0]) == str:
        str_vector = agg_vector
        agg_vector = np.zeros(len(str_vector))
        if pos_dict:
            if len(pos_dict.keys()) != len(set(str_vector)):
                raise ValueError(
                    "Posistion elements inconsistent with aggregation vector"
                )
            seen = pos_dict
        else:
            seen = {}
        counter = 0
        for ind, item in enumerate(str_vector):
            if item not in seen:
                seen[item] = counter
                counter += 1
            agg_vector[ind] = seen[item]

    agg_vector = np.array(agg_vector, dtype=int)
    agg_vector = agg_vector.reshape((1, -1))
    row_corr = agg_vector
    col_corr = np.arange(agg_vector.size)

    agg_matrix = np.zeros((row_corr.max() + 1, col_corr.max() + 1))
    agg_matrix[row_corr, col_corr] = 1

    # set columns with -1 value to 0
    agg_matrix[np.tile(agg_vector == -1, (np.shape(agg_matrix)[0], 1))] = 0

    return agg_matrix


def diagonalize_columns_to_sectors(
    df: pd.DataFrame, sector_index_level: Union[str, int] = "sector"
) -> pd.DataFrame:
    """Adds the resolution of the rows to columns by diagonalizing

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to diagonalize
    sector_index_name : string, optional
        Name or number of the index level containing sectors.

    Returns
    -------
    pd.DataFrame, diagonalized


    Example
    --------
        input       output
         (all letters are index or header)
            A B     A A A B B B
                    x y z x y z
        A x 3 1     3 0 0 1 0 0
        A y 4 2     0 4 0 0 2 0
        A z 5 3     0 0 5 0 0 3
        B x 6 9     6 0 0 9 0 0
        B y 7 6     0 7 0 0 6 0
        B z 8 4     0 0 8 0 0 4

    """

    sectors = df.index.get_level_values(sector_index_level).unique()
    sector_name = sector_index_level if type(sector_index_level) is str else "sector"

    new_col_index = [
        tuple(list(orig) + [new]) for orig in df.columns for new in sectors
    ]

    diag_df = pd.DataFrame(
        data=diagonalize_blocks(df.values, blocksize=len(sectors)),
        index=df.index,
        columns=pd.MultiIndex.from_product(
            [df.columns, sectors], names=[*df.columns.names, sector_name]
        ),
    )
    return diag_df


def diagonalize_blocks(arr: np.array, blocksize: int):
    """Diagonalize sections of columns of an array for the whole array

    Parameters
    ----------

    arr : numpy array
        Input array

    blocksize : int
        number of rows/colums forming one block

    Returns
    -------
    numpy ndarray with shape (columns 'arr' * blocksize,
                              columns 'arr' * blocksize)

    Example
    --------

    arr:      output: (blocksize = 3)
        3 1     3 0 0 1 0 0
        4 2     0 4 0 0 2 0
        5 3     0 0 5 0 0 3
        6 9     6 0 0 9 0 0
        7 6     0 7 0 0 6 0
        8 4     0 0 8 0 0 4
    """

    nr_col = arr.shape[1]
    nr_row = arr.shape[0]

    if np.mod(nr_row, blocksize):
        raise ValueError(
            "Number of rows of input array must be a multiple of blocksize"
        )

    arr_diag = np.zeros((nr_row, blocksize * nr_col))

    for col_ind, col_val in enumerate(arr.T):
        col_start = col_ind * blocksize
        col_end = blocksize + col_ind * blocksize
        for _ind in range(int(nr_row / blocksize)):
            row_start = _ind * blocksize
            row_end = blocksize + _ind * blocksize
            arr_diag[row_start:row_end, col_start:col_end] = np.diag(
                col_val[row_start:row_end]
            )

    return arr_diag


def set_dom_block(df: pd.DataFrame, value: float = 0) -> pd.DataFrame:
    """Set domestic blocks to value (0 by default)

    Requires that the region is the top index in the multiindex
    hierarchy (default case in pymrio).

    Parameters
    ----------
    df : pd.DataFrame
    value : float, optional

    Returns
    -------
    pd.DataFrame

    """
    regions = df.index.get_level_values(0).unique()
    df_res = df.copy()
    for reg in regions:
        df_res.loc[reg, reg] = 0
    return df_res


def set_block(arr, arr_block):
    """Sets the diagonal blocks of an array to an given array

    Parameters
    ----------
    arr : numpy ndarray
        the original array
    block_arr : numpy ndarray
        the block array for the new diagonal

    Returns
    -------
    numpy ndarray (the modified array)

    """
    nr_col = arr.shape[1]
    nr_row = arr.shape[0]

    nr_col_block = arr_block.shape[1]
    nr_row_block = arr_block.shape[0]

    if np.mod(nr_row, nr_row_block) or np.mod(nr_col, nr_col_block):
        raise ValueError(
            "Number of rows/columns of the input array "
            "must be a multiple of block shape"
        )
    if nr_row / nr_row_block != nr_col / nr_col_block:
        raise ValueError(
            "Block array can not be filled as " "diagonal blocks in the given array"
        )

    arr_out = arr.copy()

    for row_ind in range(int(nr_row / nr_row_block)):
        row_start = row_ind * nr_row_block
        row_end = nr_row_block + nr_row_block * row_ind
        col_start = row_ind * nr_col_block
        col_end = nr_col_block + nr_col_block * row_ind

        arr_out[row_start:row_end, col_start:col_end] = arr_block

    return arr_out


def unique_element(ll):
    """returns unique elements from a list preserving the original order"""
    seen = {}
    result = []
    for item in ll:
        if item in seen:
            continue
        seen[item] = 1
        result.append(item)
    return result


def build_agg_vec(agg_vec, **source):
    """Builds an combined aggregation vector based on various classifications

    This function build an aggregation vector based on the order in agg_vec.
    The naming and actual mapping is given in source, either explicitly or by
    pointing to a folder with the mapping.

    >>> build_agg_vec(['EU', 'OECD'], path = 'test')
    ['EU', 'EU', 'EU', 'OECD', 'REST', 'REST']

    >>> build_agg_vec(['OECD', 'EU'], path = 'test', miss='RoW')
    ['OECD', 'EU', 'OECD', 'OECD', 'RoW', 'RoW']

    >>> build_agg_vec(['EU', 'orig_regions'], path = 'test')
    ['EU', 'EU', 'EU', 'reg4', 'reg5', 'reg6']

    >>> build_agg_vec(['supreg1', 'other'], path = 'test',
    >>>        other = [None, None, 'other1', 'other1', 'other2', 'other2'])
    ['supreg1', 'supreg1', 'other1', 'other1', 'other2', 'other2']


    Parameters
    ----------
    agg_vec : list
        A list of sector or regions to which the IOSystem shall be aggregated.
        The order in agg_vec is important:
        If a string was assigned to one specific entry it will not be
        overwritten if it is given in the next vector, e.g.  ['EU', 'OECD']
        would aggregate first into EU and the remaining one into OECD, whereas
        ['OECD', 'EU'] would first aggregate all countries into OECD and than
        the remaining countries into EU.

    source : list or string
        Definition of the vectors in agg_vec.  The input vectors (either in the
        file or given as list for the entries in agg_vec) must be as long as
        the desired output with a string for every position which should be
        aggregated and None for position which should not be used.

        Special keywords:

            - path : Path to a folder with concordance matrices.
                     The files in the folder can have any extension but must be
                     in text format (tab separated) with one entry per row.
                     The last column in the file will be taken as aggregation
                     vectors (other columns can be used for documentation).
                     Values must be given for every entry in the original
                     classification (string None for all values not used) If
                     the same entry is given in source and as text file in
                     path than the one in source will be used.

                     Two special path entries are available so far:

                     - 'exio2'
                       Concordance matrices for EXIOBASE 2.0
                     - 'test'
                       Concordance matrices for the test IO system

                     If a entry is not found in source and no path is given
                     the current directory will be searched for the definition.

            - miss : Entry to use for missing values, default: 'REST'

    Returns
    -------
    list (aggregation vector)

    """

    # build a dict with aggregation vectors in source and folder
    # TODO: the logic here should be moved to constants
    if type(agg_vec) is str:
        agg_vec = [agg_vec]
    agg_dict = dict()
    for entry in agg_vec:
        try:
            agg_dict[entry] = source[entry]
        except KeyError:
            folder = source.get("path", "./")
            folder = os.path.join(PYMRIO_PATH[folder], "concordance")
            for file in os.listdir(folder):
                if entry == os.path.splitext(file)[0]:
                    _tmp = np.genfromtxt(os.path.join(folder, file), dtype=str)
                    if _tmp.ndim == 1:
                        agg_dict[entry] = [
                            None if ee == "None" else ee for ee in _tmp.tolist()
                        ]
                    else:
                        agg_dict[entry] = [
                            None if ee == "None" else ee for ee in _tmp[:, -1].tolist()
                        ]
                    break
            else:
                logging.error(
                    "Aggregation vector -- {} -- not found".format(str(entry))
                )

    # build the summary aggregation vector
    def _rep(ll, ii, vv):
        ll[ii] = vv

    miss_val = source.get("miss", "REST")

    vec_list = [agg_dict[ee] for ee in agg_vec]
    out = [
        None,
    ] * len(vec_list[0])
    for currvec in vec_list:
        if len(currvec) != len(out):
            logging.warn("Inconsistent vector length")
        [_rep(out, ind, val) for ind, val in enumerate(currvec) if not out[ind]]

    [_rep(out, ind, miss_val) for ind, val in enumerate(out) if not val]

    return out


def find_first_number(ll):
    """Returns nr of first entry parseable to float in ll, None otherwise"""
    for nr, entry in enumerate(ll):
        try:
            float(entry)
        except (ValueError, TypeError) as e:
            pass
        else:
            return nr
    return None


def sniff_csv_format(
    csv_file,
    potential_sep=["\t", ",", ";", "|", "-", "_"],
    max_test_lines=10,
    zip_file=None,
):
    """Tries to get the separator, nr of index cols and header rows in a csv file

    Parameters
    ----------

    csv_file: str
        Path to a csv file

    potential_sep: list, optional
        List of potential separators (delimiters) to test.
        Default: '\t', ',', ';', '|', '-', '_'

    max_test_lines: int, optional
        How many lines to test, default: 10 or available lines in csv_file

    zip_file: str, optional
        Path to a zip file containing the csv file (if any, default: None).
        If a zip file is given, the path given at 'csv_file' is assumed
        to be the path to the file within the zip_file.

    Returns
    -------
        dict with
            sep: string (separator)
            nr_index_col: int
            nr_header_row: int

        Entries are set to None if inconsistent information in the file
    """

    def read_first_lines(filehandle):
        lines = []
        for i in range(max_test_lines):
            line = ff.readline()
            if line == "":
                continue
            try:
                line = line.decode("utf-8")
            except AttributeError:
                pass
            lines.append(line[:-1])
        return lines

    if zip_file:
        with zipfile.ZipFile(zip_file, "r") as zz:
            with zz.open(csv_file, "r") as ff:
                test_lines = read_first_lines(ff)
    else:
        with open(csv_file, "r") as ff:
            test_lines = read_first_lines(ff)

    sep_aly_lines = [
        sorted(
            [(line.count(sep), sep) for sep in potential_sep if line.count(sep) > 0],
            key=lambda x: x[0],
            reverse=True,
        )
        for line in test_lines
    ]

    for nr, (count, sep) in enumerate(sep_aly_lines[0]):
        for line in sep_aly_lines:
            if line[nr][0] == count:
                break
        else:
            sep = None

        if sep:
            break

    lines_with_sep = [line for line in test_lines if sep in line]

    nr_header_row = None
    nr_index_col = None

    if sep:
        nr_index_col = find_first_number(lines_with_sep[-1].split(sep))
        if nr_index_col:
            for nr_header_row, line in enumerate(lines_with_sep):
                if find_first_number(line.split(sep)) == nr_index_col:
                    break

    return dict(sep=sep, nr_header_row=nr_header_row, nr_index_col=nr_index_col)


def filename_from_url(url):
    """
    Extract a file name from the download link for that file

    Parameters
    ----------

    url: str,
        The download link of the file

    Returns
    -------
    str,
        The extracted file name

    """
    name = re.search("[^/\\&\?]+\.\w{2,7}(?=([\?&].*$|$))", url)
    return name.group()


def ssl_fix(*args, **kwargs):
    """
    Tries to use a request connection with Lagacy option
    when normal connection fails

    Parameters
    ----------
    Parameters of a normal requests.get() function
        url: URL for the new :class:`Request` object.
        params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the class `Request`.
        **kwargs: Optional arguments that `request` takes.

    Returns
    -------
        r: class:`Response <Response>` object
    """

    class CustomHttpAdapter(requests.adapters.HTTPAdapter):
        # "Transport adapter" that allows us to use custom ssl_context.

        def __init__(self, ssl_context=None, **kwargs):
            self.ssl_context = ssl_context
            super().__init__(**kwargs)

        def init_poolmanager(self, connections, maxsize, block=False):
            self.poolmanager = urllib3.poolmanager.PoolManager(
                num_pools=connections,
                maxsize=maxsize,
                block=block,
                ssl_context=self.ssl_context,
            )

    try:
        r = requests.get(*args, **kwargs)
    except:
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT
        session = requests.session()
        session.mount("https://", CustomHttpAdapter(ctx))
        r = session.get(*args, **kwargs)

    return r
